/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Written by Mark Hemment, 1996 (markhe@nextd.demon.co.uk).
 *
 * (C) SGI 2006, Christoph Lameter
 * 	Cleaned up and restructured to ease the addition of alternative
 * 	implementations of SLAB allocators.
 * (C) Linux Foundation 2008-2013
 *      Unified interface for all slab allocators
 */

#ifndef _LINUX_SLAB_H
#define	_LINUX_SLAB_H

#include <linux/gfp.h>
#include <linux/overflow.h>
#include <linux/types.h>
#include <linux/workqueue.h>
#include <linux/percpu-refcount.h>
#include <linux/autoslab.h>

#include <linux/err.h>

/*
 * Flags to pass to kmem_cache_create().
 * The ones marked DEBUG are only valid if CONFIG_DEBUG_SLAB is set.
 */
/* PaX: Do not sanitize objs on free */
#define SLAB_NO_SANITIZE	((slab_flags_t __force)0x00000001U)

#ifdef CONFIG_PAX_USERCOPY
/* PaX: Allow copying objs to/from userland */
#define SLAB_USERCOPY		((slab_flags_t __force)0x00000002U)
#else
#define SLAB_USERCOPY		((slab_flags_t __force)0x00000000U)
#endif

#ifdef CONFIG_PAX_EXACT_USERCOPY
/* PaX: Precise USERCOPY checks for kmalloc */
#define SLAB_EXACT_USERCOPY	((slab_flags_t __force)0x00000004U)
#else
/* PaX: Precise USERCOPY checks for kmalloc */
#define SLAB_EXACT_USERCOPY	((slab_flags_t __force)0x00000000U)
#endif

#ifdef CONFIG_PAX_KERNSEAL
#define SLAB_SEALED		((slab_flags_t __force)0x00000008U)
#define SLAB_HIDDEN		((slab_flags_t __force)0x00000010U)
#else
#define SLAB_SEALED		0
#define SLAB_HIDDEN		0
#endif

/* DEBUG: Perform (expensive) checks on alloc/free */
#define SLAB_CONSISTENCY_CHECKS	((slab_flags_t __force)0x00000100U)
/* DEBUG: Red zone objs in a cache */
#define SLAB_RED_ZONE		((slab_flags_t __force)0x00000400U)
/* DEBUG: Poison objects */
#define SLAB_POISON		((slab_flags_t __force)0x00000800U)
/* Align objs on cache lines */
#define SLAB_HWCACHE_ALIGN	((slab_flags_t __force)0x00002000U)
/* Use GFP_DMA memory */
#define SLAB_CACHE_DMA		((slab_flags_t __force)0x00004000U)
/* Use GFP_DMA32 memory */
#define SLAB_CACHE_DMA32	((slab_flags_t __force)0x00008000U)
/* DEBUG: Store the last owner for bug hunting */
#define SLAB_STORE_USER		((slab_flags_t __force)0x00010000U)
/* The following flags affect the page allocator grouping pages by mobility */
/* Objects are reclaimable */
#define SLAB_RECLAIM_ACCOUNT	((slab_flags_t __force)0x00020000U)
#define SLAB_TEMPORARY		SLAB_RECLAIM_ACCOUNT	/* Objects are short-lived */
/* Panic if kmem_cache_create() fails */
#define SLAB_PANIC		((slab_flags_t __force)0x00040000U)
/*
 * SLAB_TYPESAFE_BY_RCU - **WARNING** READ THIS!
 *
 * This delays freeing the SLAB page by a grace period, it does _NOT_
 * delay object freeing. This means that if you do kmem_cache_free()
 * that memory location is free to be reused at any time. Thus it may
 * be possible to see another object there in the same RCU grace period.
 *
 * This feature only ensures the memory location backing the object
 * stays valid, the trick to using this is relying on an independent
 * object validation pass. Something like:
 *
 *  rcu_read_lock()
 * again:
 *  obj = lockless_lookup(key);
 *  if (obj) {
 *    if (!try_get_ref(obj)) // might fail for free objects
 *      goto again;
 *
 *    if (obj->key != key) { // not the object we expected
 *      put_ref(obj);
 *      goto again;
 *    }
 *  }
 *  rcu_read_unlock();
 *
 * This is useful if we need to approach a kernel structure obliquely,
 * from its address obtained without the usual locking. We can lock
 * the structure to stabilize it and check it's still at the given address,
 * only if we can be sure that the memory has not been meanwhile reused
 * for some other kind of object (which our subsystem's lock might corrupt).
 *
 * rcu_read_lock before reading the address, then rcu_read_unlock after
 * taking the spinlock within the structure expected at that address.
 *
 * Note that SLAB_TYPESAFE_BY_RCU was originally named SLAB_DESTROY_BY_RCU.
 */
/* Defer freeing slabs to RCU */
#define SLAB_TYPESAFE_BY_RCU	((slab_flags_t __force)0x00080000U)
/* Spread some memory over cpuset */
#define SLAB_MEM_SPREAD		((slab_flags_t __force)0x00100000U)
/* Trace allocations and frees */
#define SLAB_TRACE		((slab_flags_t __force)0x00200000U)

/* Flag to prevent checks on free */
#ifdef CONFIG_DEBUG_OBJECTS
# define SLAB_DEBUG_OBJECTS	((slab_flags_t __force)0x00400000U)
#else
# define SLAB_DEBUG_OBJECTS	0
#endif

/* Avoid kmemleak tracing */
#define SLAB_NOLEAKTRACE	((slab_flags_t __force)0x00800000U)

/* Fault injection mark */
#ifdef CONFIG_FAILSLAB
# define SLAB_FAILSLAB		((slab_flags_t __force)0x02000000U)
#else
# define SLAB_FAILSLAB		0
#endif
/* Account to memcg */
#ifdef CONFIG_MEMCG_KMEM
# define SLAB_ACCOUNT		((slab_flags_t __force)0x04000000U)
#else
# define SLAB_ACCOUNT		0
#endif

#ifdef CONFIG_KASAN
#define SLAB_KASAN		((slab_flags_t __force)0x08000000U)
#else
#define SLAB_KASAN		0
#endif

#ifdef CONFIG_GRKERNSEC_SLAB_OBJREUSE_HARDEN
/* PaX: random padding between objects */
#define SLAB_RAND_PAD		((slab_flags_t __force)0x10000000U)
#else
#define SLAB_RAND_PAD		0
#endif
/*
 * Beware that SLAB and SLUB use the remaining bits two bits 0x4000000000
 * and 0x8000000000 for internal purposes!
 */

/*
 * ZERO_SIZE_PTR will be returned for zero sized kmalloc requests.
 *
 * Dereferencing ZERO_SIZE_PTR will lead to a distinct access fault.
 *
 * ZERO_SIZE_PTR can be passed to kfree though in the same way that NULL can.
 * Both make kfree a no-op.
 */
#define ZERO_SIZE_PTR		((void *)(-MAX_ERRNO-1L))
static_assert(MAX_ERRNO & ~PAGE_MASK);

#define ZERO_OR_NULL_PTR(x) ((unsigned long)(x) - 1 >= (unsigned long)ZERO_SIZE_PTR - 1)

#include <linux/kasan.h>

struct mem_cgroup;
/*
 * struct kmem_cache related prototypes
 */
void __init kmem_cache_init(void);
bool slab_is_available(void);

extern bool usercopy_fallback;

struct kmem_cache *kmem_cache_create_usercopy_typename_print(
			const char *name, const char *typename,
			unsigned int size, unsigned int align,
			slab_flags_t flags,
			unsigned int useroffset, unsigned int usersize,
			void (*ctor)(void *),
			ssize_t (*print)(char *buffer, size_t buflen, void *obj, const char *name));

#define kmem_cache_create_usercopy_typename(name, typename, size, align, flags, useroffset, usersize, ctor) \
	kmem_cache_create_usercopy_typename_print(name, typename, size, align, flags, useroffset, usersize, ctor, NULL) \

#define kmem_cache_create_usercopy(name, size, align, flags, useroffset, usersize, ctor) \
	kmem_cache_create_usercopy_typename(name, "\0" __stringify(size), size, align, flags, useroffset, usersize, ctor)

#define kmem_cache_create(name, size, align, flags, ctor) \
	kmem_cache_create_usercopy_typename(name, "\0" __stringify(size), size, align, flags, 0, ((flags) & SLAB_USERCOPY) ? size : 0, ctor)

int kmem_cache_destroy(struct kmem_cache *s);
int kmem_cache_shrink(struct kmem_cache *s);

#if defined(AUTOSLAB_PLUGIN) && !defined(CONFIG_KASAN)
void autoslab_shrink_caches(void);
#else
/* PaX: avoid boot-time cache shrinking due to its performance impact under KASAN */
static inline void autoslab_shrink_caches(void) { }
#endif

#ifdef AUTOSLAB_PLUGIN
void __autoslab_shrink_caches(union autoslab_caches *autoslabs_start,
			      union autoslab_caches *autoslabs_stop,
			      union autoslab_caches *autoslabs_init_start,
			      union autoslab_caches *autoslabs_init_stop,
			      const char *name);
#endif

/*
 * Please use this macro to create slab caches. Simply specify the
 * name of the structure and maybe some flags that are listed above.
 *
 * The alignment of the struct determines object alignment. If you
 * f.e. add ____cacheline_aligned_in_smp to the struct declaration
 * then the objects will be properly aligned in SMP configurations.
 */
#define KMEM_CACHE(__struct, __flags)					\
		kmem_cache_create_usercopy_typename(#__struct, #__struct,\
			sizeof(struct __struct),			\
			__alignof__(struct __struct), (__flags), 0, 0, NULL)

/*
 * To whitelist a single field for copying to/from usercopy, use this
 * macro instead for KMEM_CACHE() above.
 */
#define KMEM_CACHE_USERCOPY(__struct, __flags, __field)			\
		kmem_cache_create_usercopy_typename(#__struct, #__struct,\
			sizeof(struct __struct),			\
			__alignof__(struct __struct), (__flags),	\
			offsetof(struct __struct, __field),		\
			sizeof_field(struct __struct, __field), NULL)

#define KMEM_CACHE_USERCOPY_PRINT(__struct, __flags, __field, __print)	\
		kmem_cache_create_usercopy_typename_print(#__struct, #__struct,\
			sizeof(struct __struct),			\
			__alignof__(struct __struct), (__flags),	\
			offsetof(struct __struct, __field),		\
			sizeof_field(struct __struct, __field), NULL, __print)

/*
 * Common kmalloc functions provided by all allocators
 */
void * __must_check krealloc(const void *objp, size_t new_size, gfp_t flags);
void kfree(const void *objp);
void kfree_sensitive(const void *objp);
size_t __ksize(const void *objp);
size_t ksize(const void *objp);
bool is_usercopy_object(const void *ptr);
#ifdef CONFIG_PRINTK
bool kmem_valid_obj(void *object);
void kmem_dump_obj(void *object);
#endif

/*
 * Some archs want to perform DMA into kmalloc caches and need a guaranteed
 * alignment larger than the alignment of a 64-bit integer.
 * Setting ARCH_KMALLOC_MINALIGN in arch headers allows that.
 */
#if defined(ARCH_DMA_MINALIGN) && ARCH_DMA_MINALIGN > 8
#define ARCH_KMALLOC_MINALIGN ARCH_DMA_MINALIGN
#define KMALLOC_MIN_SIZE ARCH_DMA_MINALIGN
#define KMALLOC_SHIFT_LOW ilog2(ARCH_DMA_MINALIGN)
#else
#define ARCH_KMALLOC_MINALIGN __alignof__(unsigned long long)
#endif

/*
 * Setting ARCH_SLAB_MINALIGN in arch headers allows a different alignment.
 * Intended for arches that get misalignment faults even for 64 bit integer
 * aligned buffers.
 */
#ifndef ARCH_SLAB_MINALIGN
#define ARCH_SLAB_MINALIGN __alignof__(unsigned long long)
#endif

/*
 * kmalloc and friends return ARCH_KMALLOC_MINALIGN aligned
 * pointers. kmem_cache_alloc and friends return ARCH_SLAB_MINALIGN
 * aligned pointers.
 */
#define __assume_kmalloc_alignment __assume_aligned(ARCH_KMALLOC_MINALIGN)
#define __assume_slab_alignment __assume_aligned(ARCH_SLAB_MINALIGN)
#define __assume_page_alignment __assume_aligned(PAGE_SIZE)

/*
 * Kmalloc array related definitions
 */

#ifdef CONFIG_SLAB
/*
 * The largest kmalloc size supported by the SLAB allocators is
 * 32 megabyte (2^25) or the maximum allocatable page order if that is
 * less than 32 MB.
 *
 * WARNING: Its not easy to increase this value since the allocators have
 * to do various tricks to work around compiler limitations in order to
 * ensure proper constant folding.
 */
#define KMALLOC_SHIFT_HIGH	((MAX_ORDER + PAGE_SHIFT - 1) <= 25 ? \
				(MAX_ORDER + PAGE_SHIFT - 1) : 25)
#define KMALLOC_SHIFT_MAX	KMALLOC_SHIFT_HIGH
#ifndef KMALLOC_SHIFT_LOW
#define KMALLOC_SHIFT_LOW	5
#endif
#endif

#ifdef CONFIG_SLUB
/*
 * SLUB directly allocates requests fitting in to an order-1 page
 * (PAGE_SIZE*2).  Larger requests are passed to the page allocator.
 */
#define KMALLOC_SHIFT_HIGH	(PAGE_SHIFT + 1)
#define KMALLOC_SHIFT_MAX	(MAX_ORDER + PAGE_SHIFT - 1)
#ifndef KMALLOC_SHIFT_LOW
#define KMALLOC_SHIFT_LOW	3
#endif
#endif

#ifdef CONFIG_SLOB
/*
 * SLOB passes all requests larger than one page to the page allocator.
 * No kmalloc array is necessary since objects of different sizes can
 * be allocated from the same page.
 */
#define KMALLOC_SHIFT_HIGH	PAGE_SHIFT
#define KMALLOC_SHIFT_MAX	(MAX_ORDER + PAGE_SHIFT - 1)
#ifndef KMALLOC_SHIFT_LOW
#define KMALLOC_SHIFT_LOW	3
#endif
#endif

/* Maximum allocatable size */
#define KMALLOC_MAX_SIZE	(1UL << KMALLOC_SHIFT_MAX)
/* Maximum size for which we actually use a slab cache */
#define KMALLOC_MAX_CACHE_SIZE	(1UL << KMALLOC_SHIFT_HIGH)
/* Maximum order allocatable via the slab allocator */
#define KMALLOC_MAX_ORDER	(KMALLOC_SHIFT_MAX - PAGE_SHIFT)

/*
 * Kmalloc subsystem.
 */
#ifndef KMALLOC_MIN_SIZE
#define KMALLOC_MIN_SIZE (1 << KMALLOC_SHIFT_LOW)
#endif

/*
 * This restriction comes from byte sized index implementation.
 * Page size is normally 2^12 bytes and, in this case, if we want to use
 * byte sized index which can represent 2^8 entries, the size of the object
 * should be equal or greater to 2^12 / 2^8 = 2^4 = 16.
 * If minimum size of kmalloc is less than 16, we use it as minimum object
 * size and give up to use byte sized index.
 */
#define SLAB_OBJ_MIN_SIZE      (KMALLOC_MIN_SIZE < 16 ? \
                               (KMALLOC_MIN_SIZE) : 16)

/*
 * Whenever changing this, take care of that kmalloc_type() and
 * create_kmalloc_caches() still work as intended.
 *
 * KMALLOC_NORMAL can contain only unaccounted objects whereas KMALLOC_CGROUP
 * is for accounted but unreclaimable and non-dma objects. All the other
 * kmem caches can have both accounted and unaccounted objects.
 */
enum kmalloc_cache_type {
	KMALLOC_NORMAL = 0,
#ifdef CONFIG_ZONE_DMA
	KMALLOC_DMA,
#endif
#ifdef CONFIG_ZONE_DMA32
	KMALLOC_DMA32,
#endif
	KMALLOC_RECLAIM,
#ifdef CONFIG_MEMCG_KMEM
	KMALLOC_CGROUP,
#endif
#ifdef CONFIG_PAX_USERCOPY
	KMALLOC_USERCOPY,
#ifdef CoNFIG_PAX_EXACT_USERCOPY
	KMALLOC_EXACT_USERCOPY,
#endif
#endif
#ifdef CONFIG_PAX_KERNSEAL
	KMALLOC_SEALED,
#endif
	NR_KMALLOC_TYPES
};

#ifndef CONFIG_SLOB
extern struct kmem_cache *
kmalloc_caches[NR_KMALLOC_TYPES][KMALLOC_SHIFT_HIGH + 1];

/*
 * Define gfp bits that should not be set for KMALLOC_NORMAL.
 */
#define KMALLOC_NOT_NORMAL_BITS					\
	(__GFP_RECLAIMABLE |					\
	(IS_ENABLED(CONFIG_ZONE_DMA)   ? __GFP_DMA : 0) |	\
	(IS_ENABLED(CONFIG_MEMCG_KMEM) ? __GFP_ACCOUNT : 0))

#ifdef CONFIG_PAX_EXACT_USERCOPY
void pax_store_exact_size(struct kmem_cache *cachep, void *obj, size_t size);
#else
static inline void pax_store_exact_size(struct kmem_cache *cachep, void *obj, size_t size) {}
#endif

static __always_inline __size_overflow(1) unsigned int __kmalloc_index(size_t size, bool size_is_constant);

static __always_inline __pure enum kmalloc_cache_type kmalloc_type(size_t size, gfp_t flags)
{
#ifndef CONFIG_ZONE_DMA
	flags &= ~__GFP_DMA;
#endif
#ifndef CONFIG_ZONE_DMA32
	flags &= ~__GFP_DMA32;
#endif
#ifndef CONFIG_PAX_USERCOPY
	flags &= ~__GFP_USERCOPY;
#elif defined(AUTOSLAB_PLUGIN)
	flags &= ~(GFP_USERCOPY | GFP_USER);
#endif
#ifndef CONFIG_PAX_KERNSEAL
	flags &= ~__GFP_SEALED;
#endif

	switch (flags & (__GFP_DMA | __GFP_DMA32 | __GFP_RECLAIMABLE | GFP_USERCOPY)) {
	/*
	 * At least one of the flags has to be set. Their priorities in
	 * decreasing order are:
	 *  1) __GFP_DMA
	 *  2) __GFP_RECLAIMABLE
	 *  3) __GFP_ACCOUNT
	 */
#ifdef CONFIG_ZONE_DMA
	case __GFP_DMA               | __GFP_RECLAIMABLE | GFP_USERCOPY:
	case __GFP_DMA               | __GFP_RECLAIMABLE:
	case __GFP_DMA                                   | GFP_USERCOPY:
	case __GFP_DMA:
		return KMALLOC_DMA;
#endif

#ifdef CONFIG_ZONE_DMA32
	case             __GFP_DMA32 | __GFP_RECLAIMABLE | GFP_USERCOPY:
	case             __GFP_DMA32 | __GFP_RECLAIMABLE:
	case             __GFP_DMA32                     | GFP_USERCOPY:
	case             __GFP_DMA32:
		return KMALLOC_DMA32;
#endif

#ifdef CONFIG_PAX_USERCOPY
	case                           __GFP_RECLAIMABLE | GFP_USERCOPY:
	case                                               GFP_USERCOPY:
		return KMALLOC_USERCOPY;
#endif

#ifdef CONFIG_PAX_KERNSEAL
	case                           __GFP_RECLAIMABLE | __GFP_SEALED:
	case                                               __GFP_SEALED:
		return KMALLOC_SEALED;
#endif

	case                           __GFP_RECLAIMABLE:
		return KMALLOC_RECLAIM;
	default:
#ifdef CONFIG_PAX_USERCOPY
		if ((flags & GFP_USER) == GFP_USER)
			return KMALLOC_USERCOPY;
#endif

#if defined(CONFIG_MEMCG_KMEM) && !defined(AUTOSLAB_PLUGIN)
		if (flags & __GFP_ACCOUNT)
			return KMALLOC_CGROUP;
#endif

#ifdef CONFIG_PAX_EXACT_USERCOPY
		if (size <= kmalloc_caches[KMALLOC_EXACT_USERCOPY][__kmalloc_index(size, true)]->usersize)
			return KMALLOC_EXACT_USERCOPY;
#endif
		return KMALLOC_NORMAL;
	}
}

/*
 * Figure out which kmalloc slab an allocation of a certain size
 * belongs to.
 * 0 = zero alloc
 * 1 =  65 .. 96 bytes
 * 2 = 129 .. 192 bytes
 * n = 2^(n-1)+1 .. 2^n
 *
 * Note: __kmalloc_index() is compile-time optimized, and not runtime optimized;
 * typical usage is via kmalloc_index() and therefore evaluated at compile-time.
 * Callers where !size_is_constant should only be test modules, where runtime
 * overheads of __kmalloc_index() can be tolerated.  Also see kmalloc_slab().
 */
static __always_inline __size_overflow(1) unsigned int __kmalloc_index(size_t size,
						    bool size_is_constant)
{
	if (!size)
		return 0;

	if (size <= KMALLOC_MIN_SIZE)
		return KMALLOC_SHIFT_LOW;

	if (KMALLOC_MIN_SIZE <= 32 && size > 64 && size <= 96)
		return 1;
	if (KMALLOC_MIN_SIZE <= 64 && size > 128 && size <= 192)
		return 2;
	if (size <=          8) return 3;
	if (size <=         16) return 4;
	if (size <=         32) return 5;
	if (size <=         64) return 6;
	if (size <=        128) return 7;
	if (size <=        256) return 8;
	if (size <=        512) return 9;
	if (size <=       1024) return 10;
	if (size <=   2 * 1024) return 11;
	if (size <=   4 * 1024) return 12;
	if (size <=   8 * 1024) return 13;
	if (size <=  16 * 1024) return 14;
	if (size <=  32 * 1024) return 15;
	if (size <=  64 * 1024) return 16;
	if (size <= 128 * 1024) return 17;
	if (size <= 256 * 1024) return 18;
	if (size <= 512 * 1024) return 19;
	if (size <= 1024 * 1024) return 20;
	if (size <=  2 * 1024 * 1024) return 21;
	if (size <=  4 * 1024 * 1024) return 22;
	if (size <=  8 * 1024 * 1024) return 23;
	if (size <=  16 * 1024 * 1024) return 24;
	if (size <=  32 * 1024 * 1024) return 25;

	if (0 && (IS_ENABLED(CONFIG_CC_IS_GCC) || CONFIG_CLANG_VERSION >= 110000)
	    && !IS_ENABLED(CONFIG_PROFILE_ALL_BRANCHES) && size_is_constant)
		BUILD_BUG_ON_MSG(1, "unexpected size in kmalloc_index()");
	else
		BUG();

	/* Will never be reached. Needed because the compiler may complain */
	return -1;
}
#define kmalloc_index(s) __kmalloc_index(s, true)
#endif /* !CONFIG_SLOB */

void *__kmalloc(size_t size, gfp_t flags) __assume_kmalloc_alignment __malloc __alloc_size(1) __alloc_gfp_flags(2);
void *kmem_cache_alloc(struct kmem_cache *s, gfp_t flags) __assume_slab_alignment __malloc __alloc_gfp_flags(2);
#ifdef AUTOSLAB_PLUGIN
void *kmem_cache_alloc_type(struct kmem_cache **, gfp_t flags) __assume_slab_alignment __malloc __alloc_gfp_flags(2);
void *kmem_cache_alloc_index(struct kmem_cache **, size_t size, gfp_t flags, unsigned int minshift, unsigned int maxshift) __assume_slab_alignment __malloc __alloc_size(2) __alloc_gfp_flags(3);
void *kmem_cache_alloc_index_type(struct kmem_cache **, size_t size, gfp_t flags, unsigned int minshift, unsigned int maxshift) __assume_slab_alignment __malloc __alloc_size(2) __alloc_gfp_flags(3);
#endif
void kmem_cache_free(struct kmem_cache *s, void *objp);

/*
 * Bulk allocation and freeing operations. These are accelerated in an
 * allocator specific way to avoid taking locks repeatedly or building
 * metadata structures unnecessarily.
 *
 * Note that interrupts must be enabled when calling these functions.
 */
void kmem_cache_free_bulk(struct kmem_cache *s, size_t size, void **p);
int kmem_cache_alloc_bulk(struct kmem_cache *s, gfp_t flags, size_t size, void **p);

/*
 * Caller must not use kfree_bulk() on memory not originally allocated
 * by kmalloc(), because the SLOB allocator cannot handle this.
 */
static __always_inline void kfree_bulk(size_t size, void **p)
{
	kmem_cache_free_bulk(NULL, size, p);
}

#ifdef CONFIG_NUMA
void *__kmalloc_node(size_t size, gfp_t flags, int node) __assume_kmalloc_alignment __malloc __alloc_size(1) __alloc_gfp_flags(2) __alloc_node(3);
void *kmem_cache_alloc_node(struct kmem_cache *, gfp_t flags, int node) __assume_slab_alignment __malloc __alloc_gfp_flags(2) __alloc_node(3);
#ifdef AUTOSLAB_PLUGIN
void *kmem_cache_alloc_node_type(struct kmem_cache **, gfp_t flags, int node) __assume_slab_alignment __malloc __alloc_gfp_flags(2) __alloc_node(3);
void *kmem_cache_alloc_node_index(struct kmem_cache **, size_t size, gfp_t flags, int node, unsigned int minshift, unsigned int maxshift) __assume_slab_alignment __malloc __alloc_size(2) __alloc_gfp_flags(3) __alloc_node(4);
void *kmem_cache_alloc_node_index_type(struct kmem_cache **, size_t size, gfp_t flags, int node, unsigned int minshift, unsigned int maxshift) __assume_slab_alignment __malloc __alloc_size(2) __alloc_gfp_flags(3) __alloc_node(4);
#endif
#else
#ifdef AUTOSLAB_PLUGIN
#define __kmalloc_node(size, flags, node) ({ (void)(node);__kmalloc((size), (flags)); })
#define kmem_cache_alloc_node(s, flags, node) ({ (void)(node); kmem_cache_alloc((s), (flags)); })
#else
static __always_inline __assume_kmalloc_alignment __malloc __alloc_size(1) __alloc_gfp_flags(2) __alloc_node(3)
void *__kmalloc_node(size_t size, gfp_t flags, int node)
{
	return __kmalloc(size, flags);
}

static __always_inline __alloc_gfp_flags(2) __alloc_node(3)
void *kmem_cache_alloc_node(struct kmem_cache *s, gfp_t flags, int node)
{
	return kmem_cache_alloc(s, flags);
}
#endif
#endif

#ifdef CONFIG_TRACING
void *kmem_cache_alloc_trace(struct kmem_cache *s, gfp_t flags, size_t size) __assume_slab_alignment __malloc __alloc_gfp_flags(2);

#ifdef CONFIG_NUMA
void *kmem_cache_alloc_node_trace(struct kmem_cache *s, gfp_t gfpflags,
				  int node, size_t size) __assume_slab_alignment
							__alloc_gfp_flags(2)
							__alloc_node(3)
							__malloc
							__alloc_size(4);
#else
static __always_inline __malloc __alloc_gfp_flags(2) __alloc_node(3) __alloc_size(4)
void *kmem_cache_alloc_node_trace(struct kmem_cache *s, gfp_t gfpflags, int node, size_t size)
{
	return kmem_cache_alloc_trace(s, gfpflags, size);
}
#endif /* CONFIG_NUMA */

#else /* CONFIG_TRACING */
static __always_inline __malloc __alloc_gfp_flags(2) __alloc_size(3)
void *kmem_cache_alloc_trace(struct kmem_cache *s, gfp_t flags, size_t size)
{
	void *ret = kmem_cache_alloc(s, flags);

	ret = kasan_kmalloc(s, ret, size, flags);
	return ret;
}

static __always_inline __malloc __alloc_gfp_flags(2) __alloc_node(3) __alloc_size(4) void *
kmem_cache_alloc_node_trace(struct kmem_cache *s, gfp_t gfpflags, int node, size_t size)
{
	void *ret = kmem_cache_alloc_node(s, gfpflags, node);

	ret = kasan_kmalloc(s, ret, size, gfpflags);
	return ret;
}
#endif /* CONFIG_TRACING */

extern void *kmalloc_order(size_t size, gfp_t flags, unsigned int order) __assume_page_alignment __malloc __alloc_size(1) __alloc_gfp_flags(2);

#ifdef CONFIG_TRACING
extern void *kmalloc_order_trace(size_t size, gfp_t flags, unsigned int order) __assume_page_alignment __malloc __alloc_size(1) __alloc_gfp_flags(2);
#else
static __always_inline __malloc __alloc_size(1) __alloc_gfp_flags(2)
void *kmalloc_order_trace(size_t size, gfp_t flags, unsigned int order)
{
	return kmalloc_order(size, flags, order);
}
#endif

static __always_inline __malloc __alloc_size(1) __alloc_gfp_flags(2)
void *kmalloc_large(size_t size, gfp_t flags)
{
	unsigned int order = get_order(size);
	return kmalloc_order_trace(size, flags, order);
}

/**
 * kmalloc - allocate memory
 * @size: how many bytes of memory are required.
 * @flags: the type of memory to allocate.
 *
 * kmalloc is the normal method of allocating memory
 * for objects smaller than page size in the kernel.
 *
 * The allocated object address is aligned to at least ARCH_KMALLOC_MINALIGN
 * bytes. For @size of power of two bytes, the alignment is also guaranteed
 * to be at least to the size.
 *
 * The @flags argument may be one of the GFP flags defined at
 * include/linux/gfp_types.h and described at
 * :ref:`Documentation/core-api/mm-api.rst <mm-api-gfp-flags>`
 *
 * The recommended usage of the @flags is described at
 * :ref:`Documentation/core-api/memory-allocation.rst <memory_allocation>`
 *
 * Below is a brief outline of the most useful GFP flags
 *
 * %GFP_KERNEL
 *	Allocate normal kernel ram. May sleep.
 *
 * %GFP_NOWAIT
 *	Allocation will not sleep.
 *
 * %GFP_ATOMIC
 *	Allocation will not sleep.  May use emergency pools.
 *
 * %GFP_HIGHUSER
 *	Allocate memory from high memory on behalf of user.
 *
 * Also it is possible to set different flags by OR'ing
 * in one or more of the following additional @flags:
 *
 * %__GFP_HIGH
 *	This allocation has high priority and may use emergency pools.
 *
 * %__GFP_NOFAIL
 *	Indicate that this allocation is in no way allowed to fail
 *	(think twice before using).
 *
 * %__GFP_NORETRY
 *	If memory is not immediately available,
 *	then give up at once.
 *
 * %__GFP_NOWARN
 *	If allocation fails, don't issue any warnings.
 *
 * %__GFP_RETRY_MAYFAIL
 *	Try really hard to succeed the allocation but fail
 *	eventually.
 */
#ifdef AUTOSLAB_PLUGIN
/* !!! keep in sync with the copies in include/linux/string.h and include/linux/kernel.h !!! */
void *kmalloc(size_t size, gfp_t flags) __assume_kmalloc_alignment __malloc __alloc_size(1) __alloc_gfp_flags(2);
#else
static __always_inline __malloc __alloc_size(1) __alloc_gfp_flags(2)
void *kmalloc(size_t size, gfp_t flags)
{
	if (__builtin_constant_p(size)) {
#ifndef CONFIG_SLOB
		unsigned int index;
#endif
		if (size > INT_MAX)
			return NULL;

		if (size > KMALLOC_MAX_CACHE_SIZE)
			return kmalloc_large(size, flags);
#ifndef CONFIG_SLOB
		index = kmalloc_index(size);

		if (!index)
			return ZERO_SIZE_PTR;

#ifdef CONFIG_PAX_EXACT_USERCOPY
		if ((size <   256 && index == kmalloc_index(size + 1)) ||
		    (size < 65536 && index == kmalloc_index(size + 2)) ||
		                     index == kmalloc_index(size + 4)) {
			struct kmem_cache *cachep = kmalloc_caches[KMALLOC_EXACT_USERCOPY][index];

			if (cachep) {
				void *obj = kmem_cache_alloc_trace(cachep, flags, size);

				pax_store_exact_size(cachep, obj, size);
				return obj;
			}
		}
#endif
		return kmem_cache_alloc_trace(
				kmalloc_caches[kmalloc_type(size, flags)][index],
				flags, size);
#endif
	}
	return __kmalloc(size, flags);
}
#endif

#ifdef AUTOSLAB_PLUGIN
#ifdef CONFIG_NUMA
void *kmalloc_node(size_t size, gfp_t flags, int node) __assume_kmalloc_alignment __malloc __alloc_size(1) __alloc_gfp_flags(2) __alloc_node(3);
#else
#define kmalloc_node(size, flags, node) ({ (void)(node);kmalloc((size), (flags)); })
#endif
#else
static __always_inline __malloc __alloc_size(1) __alloc_gfp_flags(2) __alloc_node(3)
void *kmalloc_node(size_t size, gfp_t flags, int node)
{
	if (__builtin_constant_p(size) && size > INT_MAX)
		return NULL;

#ifndef CONFIG_SLOB
	if (__builtin_constant_p(size) &&
		size <= KMALLOC_MAX_CACHE_SIZE) {
		unsigned int i = kmalloc_index(size);

		if (!i)
			return ZERO_SIZE_PTR;

#ifdef CONFIG_PAX_EXACT_USERCOPY
		if ((size <   256 && i == kmalloc_index(size + 1)) ||
		    (size < 65536 && i == kmalloc_index(size + 2)) ||
		                     i == kmalloc_index(size + 4)) {
			struct kmem_cache *cachep = kmalloc_caches[KMALLOC_EXACT_USERCOPY][i];

			if (cachep) {
				void *obj = kmem_cache_alloc_node_trace(cachep, flags, node, size);

				pax_store_exact_size(cachep, obj, size);
				return obj;
			}
		}
#endif
		return kmem_cache_alloc_node_trace(
				kmalloc_caches[kmalloc_type(size, flags)][i],
						flags, node, size);
	}
#endif
	return __kmalloc_node(size, flags, node);
}
#endif

/**
 * kmalloc_array - allocate memory for an array.
 * @n: number of elements.
 * @size: element size.
 * @flags: the type of memory to allocate (see kmalloc).
 */
#ifdef AUTOSLAB_PLUGIN
#define kmalloc_array(n, size, autoslab_flags)				\
({									\
	size_t __n = (n);						\
	size_t __size = (size);						\
	size_t bytes;							\
									\
	(unlikely(check_mul_overflow((size_t)(__n), (size_t)(__size), &bytes))) ?\
		NULL :							\
		kmalloc_typename(bytes, (autoslab_flags), "\0" __stringify(size));\
})
#else
static __always_inline __alloc __alloc_size(1, 2) __alloc_gfp_flags(3)
void *kmalloc_array(size_t n, size_t size, gfp_t flags)
{
	size_t bytes;

	if (unlikely(check_mul_overflow(n, size, &bytes)))
		return NULL;
	if (bytes > INT_MAX)
		return NULL;
	return kmalloc(bytes, flags);
}
#endif

/**
 * krealloc_array - reallocate memory for an array.
 * @p: pointer to the memory chunk to reallocate
 * @new_n: new number of elements to alloc
 * @new_size: new size of a single member of the array
 * @flags: the type of memory to allocate (see kmalloc)
 */
static __always_inline __alloc_size(2, 3) __must_check
void *krealloc_array(void *p, size_t new_n, size_t new_size, gfp_t flags)
{
	size_t bytes;

	if (unlikely(check_mul_overflow(new_n, new_size, &bytes)))
		return NULL;

	return krealloc(p, bytes, flags);
}

/**
 * kcalloc - allocate memory for an array. The memory is set to zero.
 * @n: number of elements.
 * @size: element size.
 * @flags: the type of memory to allocate (see kmalloc).
 */
#ifdef AUTOSLAB_PLUGIN
#define kcalloc(n, size, flags) kmalloc_array((n), (size), (flags) | __GFP_ZERO)
#else
static __always_inline __malloc __alloc_size(1, 2) __alloc_gfp_flags(3)
void *kcalloc(size_t n, size_t size, gfp_t flags)
{
	return kmalloc_array(n, size, flags | __GFP_ZERO);
}
#endif

/*
 * kmalloc_track_caller is a special version of kmalloc that records the
 * calling function of the routine calling it for slab leak tracking instead
 * of just the calling function (confusing, eh?).
 * It's useful when the call to kmalloc comes from a widely-used standard
 * allocator where we care about the real place the memory allocation
 * request comes from.
 */
extern void *__kmalloc_track_caller(size_t size, gfp_t flags, unsigned long caller) __malloc __alloc_size(1) __alloc_gfp_flags(2);
#define kmalloc_track_caller(size, flags) \
	__kmalloc_track_caller(size, flags, _RET_IP_)

#ifdef AUTOSLAB_PLUGIN
#define kmalloc_array_node(n, size, autoslab_flags, node) ({		\
	size_t bytes;							\
									\
	(unlikely(check_mul_overflow((size_t)(n), (size_t)(size), &bytes))) ?\
		NULL :							\
		kmalloc_node(bytes, (autoslab_flags), (node));		\
})
#else
static __always_inline __alloc __alloc_size(1, 2) __alloc_gfp_flags(3) __alloc_node(4)
void *kmalloc_array_node(size_t n, size_t size, gfp_t flags, int node)
{
	size_t bytes;

	if (unlikely(check_mul_overflow(n, size, &bytes)))
		return NULL;
	if (__builtin_constant_p(n) && __builtin_constant_p(size))
		return kmalloc_node(bytes, flags, node);
	return __kmalloc_node(bytes, flags, node);
}
#endif

#ifdef AUTOSLAB_PLUGIN
#define kcalloc_node(n, size, flags, node) kmalloc_array_node((n), (size), (flags) | __GFP_ZERO, (node))
#else
static __always_inline __malloc __alloc_size(1, 2) __alloc_gfp_flags(3) __alloc_node(4)
void *kcalloc_node(size_t n, size_t size, gfp_t flags, int node)
{
	return kmalloc_array_node(n, size, flags | __GFP_ZERO, node);
}
#endif


#ifdef CONFIG_NUMA
extern void *__kmalloc_node_track_caller(size_t size, gfp_t flags, int node, unsigned long caller) __malloc __alloc_size(1) __alloc_gfp_flags(2) __alloc_node(3);
#define kmalloc_node_track_caller(size, flags, node) \
	__kmalloc_node_track_caller(size, flags, node, \
			_RET_IP_)

#else /* CONFIG_NUMA */

#define kmalloc_node_track_caller(size, flags, node) \
	kmalloc_track_caller(size, flags)

#endif /* CONFIG_NUMA */

/*
 * Shortcuts
 */
static inline void *kmem_cache_zalloc(struct kmem_cache *k, gfp_t flags)
{
	return kmem_cache_alloc(k, flags | __GFP_ZERO);
}

static inline void *kmem_cache_zalloc_node(struct kmem_cache *k, gfp_t flags, int node)
{
	return kmem_cache_alloc_node(k, flags | __GFP_ZERO, node);
}

/**
 * kzalloc - allocate memory. The memory is set to zero.
 * @size: how many bytes of memory are required.
 * @flags: the type of memory to allocate (see kmalloc).
 */
#ifdef AUTOSLAB_PLUGIN
#define kzalloc(size, flags) kmalloc((size), (flags) | __GFP_ZERO)
#else
static __always_inline __malloc __alloc_size(1) __alloc_gfp_flags(2)
void *kzalloc(size_t size, gfp_t flags)
{
	return kmalloc(size, flags | __GFP_ZERO);
}
#endif

/**
 * kzalloc_node - allocate zeroed memory from a particular memory node.
 * @size: how many bytes of memory are required.
 * @flags: the type of memory to allocate (see kmalloc).
 * @node: memory node from which to allocate
 */
#ifdef AUTOSLAB_PLUGIN
#define kzalloc_node(size, flags, node) kmalloc_node((size), (flags) | __GFP_ZERO, (node))
#else
static __always_inline __malloc __alloc_size(1) __alloc_gfp_flags(2) __alloc_node(3)
void *kzalloc_node(size_t size, gfp_t flags, int node)
{
	return kmalloc_node(size, flags | __GFP_ZERO, node);
}
#endif

unsigned int kmem_cache_size(const struct kmem_cache *s);
unsigned int kmem_cache_pad_space(const struct kmem_cache *s);
const char *kmem_cache_name(const struct kmem_cache *s);
const char *kmem_cache_name_safe(const struct kmem_cache *s);
const void *kmem_cache_offset(struct kmem_cache **s, struct page *page, const void *ptr);
void __init kmem_cache_init_late(void);

#ifdef CONFIG_KALLSYMS
ssize_t kmem_cache_print(struct kmem_cache *s, char *buffer, size_t buflen, void *obj, const char *name);
void *__kmem_cache_lookup(struct kmem_cache **s, void *ptr);
const char *kmem_cache_lookup(void *ptr,
			      unsigned long *size,
			      unsigned long *offset,
			      char *buffer,
			      char **modname);
#endif

#if defined(CONFIG_SMP) && defined(CONFIG_SLAB)
int slab_prepare_cpu(unsigned int cpu);
int slab_dead_cpu(unsigned int cpu);
#else
#define slab_prepare_cpu	NULL
#define slab_dead_cpu		NULL
#endif

#ifdef CONFIG_PAX_AUTOSLAB_PLUGIN_AUTOSTACK
void *convert_autostack(void *autostack, size_t size, gfp_t flags) __maybe_used;
void free_autostack(const void *autostack, size_t size) __maybe_used;
#endif

#endif	/* _LINUX_SLAB_H */
