/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _LINUX_SLAB_DEF_H
#define	_LINUX_SLAB_DEF_H

#include <linux/kfence.h>
#include <linux/reciprocal_div.h>

/*
 * Definitions unique to the original Linux SLAB allocator.
 */

struct kmem_cache {
	/* PaX: adjust PERCPU_AUTOSLAB_RESERVE for the __percpu fields */
	struct array_cache __percpu *cpu_cache;

/* 1) Cache tunables. Protected by slab_mutex */
	unsigned int batchcount;
	unsigned int limit;
	unsigned int shared;

	unsigned int size;
	struct reciprocal_value reciprocal_buffer_size;
#ifdef CONFIG_GRKERNSEC_SLAB_OBJREUSE_HARDEN
	unsigned short pad_space;
#endif

/* 2) touched by every alloc & free from the backend */

	slab_flags_t flags;		/* constant flags */
	unsigned int num;		/* # of objs per slab */

/* 3) cache_grow/shrink */
	/* order of pgs per slab (2^n) */
	unsigned int gfporder;

	/* force GFP flags, e.g. GFP_DMA */
	gfp_t allocflags;

	size_t colour;			/* cache colouring range */
	unsigned int colour_off;	/* colour offset */
	struct kmem_cache *freelist_cache;
	unsigned int freelist_size;

	/* constructor func */
	void (*ctor)(void *obj);
#ifdef CONFIG_KALLSYMS
	ssize_t (*print)(char *buffer, size_t buflen, void *obj, const char *name);
#endif

/* 4) cache creation/removal */
	const char *name;
	struct list_head list;
	atomic_t refcount;
	int object_size;
	int align;

/* 5) statistics */
#ifdef CONFIG_DEBUG_SLAB
	unsigned long num_active;
	unsigned long num_allocations;
	unsigned long high_mark;
	unsigned long grown;
	unsigned long reaped;
	unsigned long errors;
	unsigned long max_freeable;
	unsigned long node_allocs;
	unsigned long node_frees;
	unsigned long node_overflow;
	atomic_unchecked_t allochit;
	atomic_unchecked_t allocmiss;
	atomic_unchecked_t freehit;
	atomic_unchecked_t freemiss;
#ifdef CONFIG_PAX_MEMORY_SANITIZE
	atomic_unchecked_t sanitized;
	atomic_unchecked_t not_sanitized;
#endif

	/*
	 * If debugging is enabled, then the allocator can add additional
	 * fields and/or padding to every object. 'size' contains the total
	 * object size including these internal fields, while 'obj_offset'
	 * and 'object_size' contain the offset to the user object and its
	 * size.
	 */
	int obj_offset;
#endif /* CONFIG_DEBUG_SLAB */

#ifdef CONFIG_KASAN
	struct kasan_cache kasan_info;
#endif

#ifdef CONFIG_SLAB_FREELIST_RANDOM
	unsigned int *random_seq;
#endif

	unsigned int useroffset;	/* USERCOPY region offset */
	unsigned int usersize;		/* USERCOPY region size */

#ifdef CONFIG_PAX_AUTOSLAB_PLUGIN_AUTOTYPENAME
	const char *typename;
#endif

	struct kmem_cache_node *node[MAX_NUMNODES];
};

static inline const void *nearest_obj(struct kmem_cache *cache, struct page *page,
				const void *x)
{
	const void *object = x - (x - page->s_mem) % cache->size;
	const void *last_object = page->s_mem + (cache->num - 1) * cache->size;

	if (unlikely(object > last_object))
		return last_object;
	else
		return object;
}

/*
 * We want to avoid an expensive divide : (offset / cache->size)
 *   Using the fact that size is a constant for a particular cache,
 *   we can replace (offset / cache->size) by
 *   reciprocal_divide(offset, cache->reciprocal_buffer_size)
 */
static inline unsigned int obj_to_index(const struct kmem_cache *cache,
					const struct page *page, const void *obj)
{
	u32 offset = (obj - page->s_mem);
	return reciprocal_divide(offset, cache->reciprocal_buffer_size);
}

static inline void *index_to_obj(struct kmem_cache *cache, struct page *page,
				 unsigned int idx)
{
	return page->s_mem + cache->size * idx;
}

#ifdef CONFIG_DEBUG_SLAB
static int obj_offset(struct kmem_cache *cachep)
{
	return cachep->obj_offset;
}
#else
#define obj_offset(x)			0
#endif

static inline const void *ptr_to_obj(struct kmem_cache *cachep, struct page *page, const void *ptr)
{
	unsigned int objnr;

	if (ptr < page->s_mem)
		return NULL;
	objnr = obj_to_index(cachep, page, ptr);
	BUG_ON(objnr >= cachep->num);

	return index_to_obj(cachep, page, objnr) + obj_offset(cachep);
}

void print_objinfo(struct kmem_cache *cachep, void *objp, int lines);

static inline int objs_per_slab_page(const struct kmem_cache *cache,
				     const struct page *page)
{
	if (is_kfence_address(page_address(page)))
		return 1;
	return cache->num;
}

#endif	/* _LINUX_SLAB_DEF_H */
