/*
 * Copyright 2011 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2011-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "constify.h"

typedef struct {
	bool has_fptr_field;
	bool has_non_integral_field;
	bool has_writable_field;
	bool has_do_const_field;
	bool has_mutable_const_field;
	bool all_mutable_const_fields;
	bool has_no_const_field;
	bool verbose;
} constify_info;

/*
 * determine whether the given structure type meets the requirements for automatic constification,
 * including the constification attributes on nested structure types
 */
static void constifiable(const_tree node, constify_info *cinfo)
{
	const_tree field;

	gcc_assert(TREE_CODE(node) == RECORD_TYPE || TREE_CODE(node) == UNION_TYPE);

	// e.g., pointer to structure fields while still constructing the structure type
	if (TYPE_FIELDS(node) == NULL_TREE)
		return;

	for (field = TYPE_FIELDS(node); field; field = TREE_CHAIN(field)) {
		const_tree type = get_inner_type(field);
		enum tree_code code = TREE_CODE(type);

		if (node == type)
			continue;

		// this is not something the frontend can produce but randomize_layout does
		if (type == void_type_node)
			continue;

		if (FUNCTION_POINTER_TYPE_P(type))
			// fptrs don't destroy the 'all mutable const' property
			cinfo->has_fptr_field = true;
		else if (code == RECORD_TYPE || code == UNION_TYPE) {
			if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type))) {
				cinfo->all_mutable_const_fields = false;
				cinfo->has_do_const_field = true;
				if (cinfo->verbose)
					inform(DECL_SOURCE_LOCATION(field), "constify: type %qT has field %qD with 'do_const' attribute", node, field);
			} else if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
				cinfo->has_mutable_const_field = true;
				cinfo->has_writable_field = true;
			} else if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type))) {
				cinfo->all_mutable_const_fields = false;
				cinfo->has_no_const_field = true;
				cinfo->has_writable_field = true;
			} else
				constifiable(type, cinfo);
		} else {
			// r/o members don't destroy the 'all mutable const' property either
			if (!TREE_READONLY(field)) {
				cinfo->has_writable_field = true;
				cinfo->all_mutable_const_fields = false;
			}
			if (!INTEGRAL_TYPE_P(type))
				cinfo->has_non_integral_field = true;
		}
	}
}

static void constify_type(tree type)
{
	tree t;

	gcc_assert(type == TYPE_MAIN_VARIANT(type));
	gcc_assert(type == TYPE_CANONICAL(type));
	TYPE_READONLY(type) = 1;
	C_TYPE_FIELDS_READONLY(type) = 1;
	TYPE_CONSTIFY_VISITED(type) = 1;

	gcc_assert(type == TYPE_CANONICAL(type));

	// fix up any already existing equivalent constified types
	for (t = TYPE_MAIN_VARIANT(type); t; t = TYPE_NEXT_VARIANT(t)) {
		tree pointer_to;

		if (t == type)
			continue;

		if (TYPE_QUALS(t) != TYPE_QUALS(type))
			continue;

		TYPE_CANONICAL(t) = TYPE_CANONICAL(type);

		if (!TYPE_POINTER_TO(t)) {
			gcc_assert(!TYPE_POINTER_TO(type));
			continue;
		}

		if (TYPE_POINTER_TO(type))
			continue;

		for (pointer_to = TYPE_POINTER_TO(t); pointer_to; pointer_to = TYPE_NEXT_PTR_TO(pointer_to))
			TREE_TYPE(pointer_to) = type;

		TYPE_POINTER_TO(type) = TYPE_POINTER_TO(t);
	}
}

static bool eligible_for_mutable_const(const constify_info *cinfo)
{
	if (cinfo->has_no_const_field || cinfo->has_do_const_field)
		return false;

	if (cinfo->has_non_integral_field)
		return false;

	if (!cinfo->has_mutable_const_field || !cinfo->all_mutable_const_fields)
		return false;

	return true;
}

void finish_type(void *event_data, void *data __unused)
{
	tree mutable_attr, type = (tree)event_data;
	constify_info cinfo = {
		.has_fptr_field = false,
		.has_non_integral_field = false,
		.has_writable_field = false,
		.has_do_const_field = false,
		.has_mutable_const_field = false,
		.all_mutable_const_fields = true,
		.has_no_const_field = false,
		.verbose = false,
	};

	if (type == NULL_TREE || type == error_mark_node)
		return;

#if BUILDING_GCC_VERSION >= 5000
	if (TREE_CODE(type) == ENUMERAL_TYPE)
		return;
#endif

	if (TYPE_FIELDS(type) == NULL_TREE || TYPE_CONSTIFY_VISITED(type))
		return;

	constifiable(type, &cinfo);

	// handle mutable_const: don't allow mixing with do_const fields
	mutable_attr = lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type));
	if (mutable_attr) {
		if (cinfo.has_do_const_field)
			error("'mutable_const' attribute used on type %qT that is partially constified", type);
		else {
			if (verbose && !lookup_attribute("mutable_const_performed", TYPE_ATTRIBUTES(type)) && eligible_for_mutable_const(&cinfo)) {
				location_t loc = TYPE_STUB_DECL(type) ? DECL_SOURCE_LOCATION(TYPE_STUB_DECL(type)) : UNKNOWN_LOCATION;
				inform(loc, "constify: redundant 'mutable_const' attribute for %qT", type);
			}
			TYPE_CONSTIFY_VISITED(type) = 1;
		}
		return;
	}

	// handle no_const: don't allow on types that wouldn't be constified anyway
	if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type))) {
		if ((cinfo.has_fptr_field && !cinfo.has_writable_field) || cinfo.has_do_const_field || cinfo.has_mutable_const_field) {
			TYPE_CONSTIFY_VISITED(type) = 1;
		} else
			error("'no_const' attribute used on type %qT that is not constified", type);
		return;
	}

	// handle do_const: don't allow on types that are eligible for automatic constification, even if they lack fptr fields
	if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type))) {
		if (!cinfo.has_writable_field) {
			error("'do_const' attribute used on type %qT that is%sautomatically constified", type, cinfo.has_fptr_field ? " " : " not ");
			return;
		}
		constify_type(type);
		return;
	}

	// propagate mutable_const from fields to outer types composed of only fptrs and other mutable fields
	if (eligible_for_mutable_const(&cinfo)) {
		TYPE_CONSTIFY_VISITED(type) = 1;
		if (cinfo.verbose) {
			location_t loc = TYPE_STUB_DECL(type) ? DECL_SOURCE_LOCATION(TYPE_STUB_DECL(type)) : UNKNOWN_LOCATION;
			inform(loc, "constify: propagating 'mutable_const' attribute to %qT", type);
		}
		add_type_attr(type, "mutable_const", NULL_TREE);
		add_type_attr(type, "mutable_const_performed", NULL_TREE);
		return;
	}

	// automatic constification of types with fptrs and immutable fields
	if ((cinfo.has_fptr_field || cinfo.has_do_const_field) && !cinfo.has_writable_field) {
		constify_type(type);
		add_type_attr(type, "do_const", NULL_TREE);
		return;
	}

	if (cinfo.has_do_const_field) {
		error("non-constifiable type %qT has field with 'do_const' attribute", type);
		cinfo.has_fptr_field = false,
		cinfo.has_non_integral_field = false,
		cinfo.has_writable_field = false,
		cinfo.has_do_const_field = false,
		cinfo.has_mutable_const_field = false,
		cinfo.has_no_const_field = false,
		cinfo.verbose = true;
		constifiable(type, &cinfo);
	}
	TYPE_CONSTIFY_VISITED(type) = 1;
}
