/*
 * Copyright 2012-2023 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

static void check_fptr_initializer(tree var, tree fptr_type, tree init_type)
{
	rap_hash_t hash_fptr, hash_func;
	tree func_type;
	bool convertible;

	if (TREE_CODE(fptr_type) != POINTER_TYPE)
		return;

	convertible = !!lookup_attribute("convertible", TYPE_ATTRIBUTES(fptr_type));

	fptr_type = TREE_TYPE(fptr_type);
	if (TREE_CODE(fptr_type) != FUNCTION_TYPE)
		return;

	if (TREE_CODE(init_type) != ADDR_EXPR)
		return;
	func_type = TREE_TYPE(init_type);
	if (TREE_CODE(func_type) != POINTER_TYPE)
		return;

	convertible = convertible || !!lookup_attribute("convertible", TYPE_ATTRIBUTES(func_type));

	func_type = TREE_TYPE(func_type);
	if (TREE_CODE(func_type) != FUNCTION_TYPE)
		return;

	hash_fptr = rap_lookup_precise_rap_type_hash(fptr_type);
	hash_func = rap_lookup_precise_rap_type_hash(func_type);
	if (hash_fptr.hash == hash_func.hash)
		return;

	if (convertible)
		return;

	error_at(DECL_SOURCE_LOCATION(var), "rap hash mismatch between variable %qD (%x) and function %qE (%x)",
		 var, hash_fptr.hash, init_type, hash_func.hash);
}

static void check_constructor(tree var, tree var_type, tree constructor)
{
	unsigned int i;
	tree index, value;

	FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(constructor), i, index, value) {
		rap_hash_t hash_fptr, hash_func;
		tree fptr_type, func_type, context;

		if (TREE_CODE(var_type) == ARRAY_TYPE) {
			fptr_type = TREE_TYPE(var_type);
			if (TREE_CODE(fptr_type) == ARRAY_TYPE) {
				if (TREE_CODE(value) == CONSTRUCTOR)
					check_constructor(var, fptr_type, value);
				continue;
			}
		} else {
			gcc_assert(RECORD_OR_UNION_TYPE_P(var_type));

			if (!index) {
				gcc_assert(!strncmp(TYPE_NAME_POINTER(TREE_TYPE(var)), "__ubsan_", 8));
				continue;
			}

			gcc_assert(TREE_CODE(index) == FIELD_DECL);
			fptr_type = TREE_TYPE(index);
		}

		// check array/struct fields recursively
		if (RECORD_OR_UNION_TYPE_P(fptr_type)) {
			if (TREE_CODE(value) == NOP_EXPR)
				value = TREE_OPERAND(value, 0);
			gcc_assert(TREE_CODE(value) == CONSTRUCTOR);
			check_constructor(var, fptr_type, value);
			continue;
		}

		if (TREE_CODE(fptr_type) != POINTER_TYPE)
			continue;

		fptr_type = TREE_TYPE(fptr_type);
		if (TREE_CODE(fptr_type) != FUNCTION_TYPE)
			continue;

		gcc_assert(TREE_CODE(value) != CONSTRUCTOR);
		if (TREE_CODE(value) == NOP_EXPR)
			value = TREE_OPERAND(value, 0);
		if (TREE_CODE(value) != ADDR_EXPR) {
			if (TREE_CODE(value) == INTEGER_CST)
				continue;

			debug_tree(index);
			debug_tree(value);
			gcc_unreachable();
			continue;
		}
		func_type = TREE_TYPE(value);
		if (TREE_CODE(func_type) != POINTER_TYPE)
			continue;
		func_type = TREE_TYPE(func_type);
		if (TREE_CODE(func_type) != FUNCTION_TYPE)
			continue;

		hash_fptr = rap_lookup_precise_rap_type_hash(fptr_type);
		hash_func = rap_lookup_precise_rap_type_hash(func_type);

		// handle initializers of unions as gcc assigns to the first field
		// whereas the actually used field may be different
		if (TREE_CODE(index) == FIELD_DECL) {
			context = DECL_FIELD_CONTEXT(index);
			gcc_assert(context);
			if (TREE_CODE(context) == UNION_TYPE) {
				tree field;

				for (field = TYPE_FIELDS(context); field; field = TREE_CHAIN(field)) {
					fptr_type = TREE_TYPE(field);
					if (TREE_CODE(fptr_type) != POINTER_TYPE)
						continue;
					fptr_type = TREE_TYPE(fptr_type);
					if (TREE_CODE(fptr_type) != FUNCTION_TYPE)
						continue;
					hash_fptr = rap_lookup_precise_rap_type_hash(fptr_type);
					if (hash_fptr.hash == hash_func.hash)
						break;
				}
			}
		}
		if (hash_fptr.hash == hash_func.hash)
			continue;

		if (TREE_CODE(index) == FIELD_DECL) {
			tree field_type;

			field_type = TREE_TYPE(index);
			gcc_assert(TREE_CODE(field_type) == POINTER_TYPE);

			if (FUNCTION_POINTER_TYPE_P(field_type))
				if (lookup_attribute("convertible", TYPE_ATTRIBUTES(TREE_TYPE(index))))
					continue;
		} else {
			if (lookup_attribute("convertible", TYPE_ATTRIBUTES(TREE_TYPE(var_type))))
				continue;
		}

		if (TREE_CODE(var_type) == ARRAY_TYPE)
			error_at(DECL_SOURCE_LOCATION(var), "rap hash mismatch between '%qD[%E]' of type %qT (%x) and function pointer %qE of type %qT (%x)",
				 var, index, TREE_TYPE(var_type), hash_fptr.hash, value, TREE_TYPE(value), hash_func.hash);
		else
			error_at(DECL_SOURCE_LOCATION(var), "rap hash mismatch between field %qD of type %qT (%x) and function pointer %qE of type %qT (%x)",
				 index, TREE_TYPE(index), hash_fptr.hash, value, TREE_TYPE(value), hash_func.hash);
	}
}

/*
 * try to deconstruct the variable initializer and verify all function pointers
 */
void check_variable_casts(tree var)
{
	tree var_type, ctor;

	if (DECL_EXTERNAL(var) || !DECL_INITIAL(var))
		return;

	ctor = DECL_INITIAL(var);
	var_type = TREE_TYPE(var);
	switch (TREE_CODE(var_type)) {
	default:
		return;

	case POINTER_TYPE:
		check_fptr_initializer(var, var_type, ctor);
		return;

	case RECORD_TYPE:
	case UNION_TYPE:
	case ARRAY_TYPE:
		if (TREE_CODE(ctor) != CONSTRUCTOR)
			return;
		check_constructor(var, var_type, ctor);
		break;
	}
}

static void check_call_args(gimple_stmt_ptr call_stmt)
{
	tree fn, param;
	unsigned int i, e;

	fn = gimple_call_fn(call_stmt);
	if (!fn || TREE_CODE(fn) == INTEGER_CST)
		return;

	switch (TREE_CODE(fn)) {
	default:
		debug_tree(fn);
		debug_tree(TREE_TYPE(fn));
		debug_tree(TREE_TYPE(TREE_TYPE(fn)));
		gcc_unreachable();

	case ADDR_EXPR:
	case PARM_DECL:
	case SSA_NAME:
	case VAR_DECL:
		break;
	}

	fn = TREE_TYPE(fn);
	gcc_assert(POINTER_TYPE_P(fn));

	fn = TREE_TYPE(fn);
	gcc_assert(TREE_CODE(fn) == FUNCTION_TYPE);

	e = gimple_call_num_args(call_stmt);

	for (i = 0, param = TYPE_ARG_TYPES(fn); param && i < e; ++i, param = TREE_CHAIN(param)) {
		tree arg, fptr, fptr_type;
		rap_hash_t hash_fptr, hash_func;

		fptr = TREE_VALUE(param);
		if (TREE_CODE(fptr) != POINTER_TYPE)
			continue;
		fptr_type = TREE_TYPE(fptr);
		if (TREE_CODE(fptr_type) != FUNCTION_TYPE)
			continue;

		arg = gimple_call_arg(call_stmt, i);
		switch (TREE_CODE(arg)) {
		default:
			debug_tree(arg);
			debug_tree(TREE_TYPE(arg));
			debug_tree(TREE_TYPE(TREE_TYPE(arg)));
			gcc_unreachable();

		case INTEGER_CST:
				continue;

		case PARM_DECL:
		case SSA_NAME:
		case VAR_DECL: {
			tree func_type = TREE_TYPE(TREE_TYPE(arg));

			hash_func = rap_lookup_precise_rap_type_hash(func_type);
			break;
		}

		case ADDR_EXPR: {
			tree func_decl = TREE_OPERAND(arg, 0);

			hash_func = rap_lookup_precise_rap_type_hash(TREE_TYPE(func_decl));
			break;
		}
		}

		hash_fptr = rap_lookup_precise_rap_type_hash(fptr_type);

		if (hash_fptr.hash == hash_func.hash)
			continue;

		if (lookup_attribute("convertible", TYPE_ATTRIBUTES(fptr)))
			continue;

		error_at(gimple_location(call_stmt), "rap hash mismatch between type %qT (%x) of param %u and type %qT (%x) of function %qE",
			 fptr, hash_fptr.hash, i + 1, TREE_TYPE(arg), hash_func.hash, arg);
	}
}

static void check_fptr_assigns(gimple_stmt_ptr cast_stmt)
{
	// gimple match: h_1 = get_fptr (); D.2709_3 = h_1 (x_2(D));
	tree lhs, rhs, fptr_type, func_type;
	rap_hash_t hash_fptr, hash_func;
	gimple_stmt_ptr def_stmt = NULL;
	location_t loc = gimple_location(cast_stmt);

	rhs = gimple_assign_rhs1(cast_stmt);
	if (TREE_CODE(rhs) == SSA_NAME) {
		def_stmt = SSA_NAME_DEF_STMT(rhs);
		if (gimple_code(def_stmt) == GIMPLE_ASSIGN)
			rhs = gimple_assign_rhs1(def_stmt);
		else {
			def_stmt = NULL;
			if (SSA_NAME_VAR(rhs))
				rhs = SSA_NAME_VAR(rhs);
		}
	}

	lhs = gimple_get_lhs(cast_stmt);

	func_type = TREE_CODE(TREE_TYPE(rhs)) == POINTER_TYPE ? TREE_TYPE(TREE_TYPE(rhs)) : NULL_TREE;
	fptr_type = TREE_CODE(TREE_TYPE(lhs)) == POINTER_TYPE ? TREE_TYPE(TREE_TYPE(lhs)) : NULL_TREE;

	if (func_type != NULL_TREE)
		func_type = TREE_CODE(func_type) == FUNCTION_TYPE ? func_type : NULL_TREE;
	if (fptr_type != NULL_TREE)
		fptr_type = TREE_CODE(fptr_type) == FUNCTION_TYPE ? fptr_type : NULL_TREE;

	if (func_type == NULL_TREE && fptr_type == NULL_TREE)
		return;

	if (func_type == NULL_TREE && fptr_type != NULL_TREE)
		return;

	if (TREE_CODE(TREE_TYPE(lhs)) != POINTER_TYPE)
		return;

	if (func_type != NULL_TREE && fptr_type == NULL_TREE)
		return;

	hash_fptr = rap_lookup_precise_rap_type_hash(fptr_type);
	hash_func = rap_lookup_precise_rap_type_hash(func_type);

	if (hash_fptr.hash == hash_func.hash) {
		gimple_stmt_ptr def_def_stmt;

#if BUILDING_GCC_VERSION >= 4006
		if (TREE_CODE(rhs) != MEM_REF)
#else
		if (TREE_CODE(rhs) != INDIRECT_REF)
#endif
			return;

		rhs = TREE_OPERAND(rhs, 0);
		if (TREE_CODE(rhs) != SSA_NAME)
			return;

		def_stmt = SSA_NAME_DEF_STMT(rhs);
		if (!is_gimple_assign(def_stmt))
			return;

		if (gimple_num_ops(def_stmt) > 2)
			return;

		rhs = gimple_assign_rhs1(def_stmt);
		if (TREE_CODE(TREE_TYPE(rhs)) != POINTER_TYPE)
			return;

		switch TREE_CODE(rhs) {
		default:
			debug_gimple_stmt(cast_stmt);
			debug_gimple_stmt(def_stmt);
			debug_tree(rhs);
			gcc_unreachable();

		case ADDR_EXPR:
		case ARRAY_REF:
		case COMPONENT_REF:
#if BUILDING_GCC_VERSION >= 4006
		case MEM_REF:
#endif
		case PARM_DECL:
		case VAR_DECL:
			return;

		case SSA_NAME:
			break;
		};

		def_def_stmt = SSA_NAME_DEF_STMT(rhs);
		if (is_gimple_assign(def_def_stmt)) {
			tree def_rhs = gimple_assign_rhs1(def_def_stmt);
			if (TREE_CODE(def_rhs) == ADDR_EXPR) {
				def_rhs = TREE_TYPE(TREE_TYPE(def_rhs));
				if (TREE_CODE(def_rhs) == ARRAY_TYPE)
					return;
			}
		}

		func_type = TREE_TYPE(TREE_TYPE(rhs));
		if (TREE_CODE(func_type) == POINTER_TYPE)
			func_type = TREE_TYPE(func_type);
		if (TREE_CODE(func_type) == FUNCTION_TYPE) {
			hash_func = rap_lookup_precise_rap_type_hash(func_type);
			if (hash_fptr.hash == hash_func.hash)
				return;
		}

		if (gimple_num_ops(def_stmt) == 2)
			return;

		error_at(gimple_location(def_stmt), "pointer %qE of type %qT cast to a different function pointer type %qT", rhs, TREE_TYPE(rhs), TREE_TYPE(lhs));
		return;
	}

	if (lookup_attribute("noderef", TYPE_ATTRIBUTES(TREE_TYPE(lhs))) ||
	    lookup_attribute("convertible", TYPE_ATTRIBUTES(TREE_TYPE(lhs))) ||
	    lookup_attribute("convertible", TYPE_ATTRIBUTES(TREE_TYPE(rhs))))
		return;

	if (TREE_CODE(rhs) == SSA_NAME) {
		gimple_stmt_ptr rhs_def_stmt;

		rhs_def_stmt = SSA_NAME_DEF_STMT(rhs);
		if (is_gimple_call(rhs_def_stmt)) {
			tree def_rhs = gimple_call_return_type(as_a_gcall(rhs_def_stmt));

			if (lookup_attribute("convertible", TYPE_ATTRIBUTES(def_rhs)))
					return;
		}
	}

	error_at(loc, "rap hash mismatch between function pointer %qE of type %qT (%x) and function pointer %qE of type %qT (%x)",
		 def_stmt ? gimple_assign_rhs1(def_stmt) : rhs, TREE_TYPE(rhs), hash_fptr.hash, lhs, TREE_TYPE(lhs), hash_func.hash);
}

static unsigned int rap_check_local_variable_casts_execute(void)
{
	tree var;
	basic_block bb;
	unsigned int i;

	FOR_EACH_LOCAL_DECL(cfun, i, var) {
		gcc_assert(DECL_P(var));
		check_variable_casts(var);
	}

	// 1. loop through BBs and GIMPLE statements
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;

			// is it a call ...
			stmt = gsi_stmt(gsi);
			if (is_gimple_call(stmt)) {
				check_call_args(stmt);
				continue;
			}

			// ... or a cast assignment
			if (is_gimple_assign(stmt) && gimple_num_ops(stmt) == 2) {
				check_fptr_assigns(stmt);
				continue;
			}
		}
	}

	return 0;
}

#define PASS_NAME rap_check_local_variable_casts
#define NO_GATE
#include "gcc-generate-gimple-pass.h"

void check_register_variable(tree var)
{
	rtx reg;

	gcc_assert(DECL_P(var));
	if (!DECL_REGISTER(var) || !DECL_ASSEMBLER_NAME_SET_P(var))
		return;

	if (!DECL_RTL(var))
		return;

	reg = DECL_RTL(var);
	if (!REG_P(reg))
		return;

	if (rap_reg_id != ORIGINAL_REGNO(reg))
		return;

	error_at(DECL_SOURCE_LOCATION(var), "variable %qD uses a reserved RAP register %s", var, rap_reg_name);
}

static unsigned int rap_check_local_register_variables_execute(void)
{
	tree var;
	unsigned int i;

	FOR_EACH_LOCAL_DECL(cfun, i, var)
		check_register_variable(var);

	return 0;
}

#define PASS_NAME rap_check_local_register_variables
#define NO_GATE
#include "gcc-generate-gimple-pass.h"
