/*
 * Copyright 2012-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

#ifdef GCC_DF_H
static bool rap_xor_elimination_gate(void)
{
	rtx_insn *insn;

	if (!optimize)
		return false;

	// handle -fdump-passes
	if (!cfun->cfg)
		return true;

	if (!rap_cmodel_check())
		return false;

	if (rap_noreturn_function())
		return false;

	if (lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl)))
		return false;

	gcc_assert(complex_functions);
	if (!bitmap_bit_p(complex_functions, NODE_UID(cgraph_get_node(current_function_decl))))
		return false;

	if (rap_calls_complex_function_p(current_function_decl))
		return false;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		rtx body;

		if (!CALL_P(insn) && !NONJUMP_INSN_P(insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);

		if (NONJUMP_INSN_P(insn)) {
			if (GET_CODE(body) == SET)
				body = SET_SRC(body);
			if (GET_CODE(body) != ASM_OPERANDS)
				continue;
			if (regexec(&asm_call_insn_regex,  ASM_OPERANDS_TEMPLATE(body), 0, NULL, 0))
				continue;
			return false;
		}

		if (SIBLING_CALL_P(insn))
			continue;

		if (GET_CODE(body) == SET)
			body = SET_SRC(body);

		if (GET_CODE(body) != CALL)
			return false;

		body = XEXP(body, 0);
		if (!MEM_P(body))
			return false;

		body = XEXP(body, 0);
		if (GET_CODE(body) != SYMBOL_REF)
			return false;
		gcc_assert(SYMBOL_REF_FUNCTION_P(body));

		if (rap_is_complex_function(SYMBOL_REF_DECL(body)))
			return false;
	}

	return true;
}

enum rap_xor { xor_none, xor_entry, xor_exit };

static enum rap_xor rap_match_xor(rtx body, bool both)
{
	rtx dst, src, src1, src2;
	tree expr;
	enum rap_xor ret;

	if (GET_CODE(body) != SET)
		return xor_none;

	dst = SET_DEST(body);
	if (!REG_P(dst) || !REG_EXPR(dst))
		return xor_none;
	expr = REG_EXPR(dst);
	if (!DECL_P(expr))
		return xor_none;
	if (!DECL_NAME(expr))
		return xor_none;
	if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_entry_", sizeof("rap_retaddr_entry_") - 1))
		ret = xor_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_exit_", sizeof("rap_retaddr_exit_") - 1))
		ret = xor_exit;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_entry", sizeof("rap_cookie_entry") - 1))
		ret = xor_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_exit", sizeof("rap_cookie_exit") - 1))
		ret = xor_exit;
	else
		return xor_none;

	src = SET_SRC(body);
	if (GET_CODE(src) != XOR)
		return xor_none;
	if (GET_MODE(dst) != GET_MODE(src))
		return xor_none;

	src1 = XEXP(src, 0);
	if (!REG_P(src1))
		return xor_none;
#if BUILDING_GCC_VERSION <= 4008
	if (!REG_EXPR(src1) || !DECL_NAME(REG_EXPR(src1)) || strncmp(DECL_NAME_POINTER(REG_EXPR(src1)), "rap_retaddr_", sizeof("rap_retaddr_") - 1))
		return xor_none;
#endif

	if (both) {
		src2 = XEXP(src, 1);

		if (!REG_P(src2) || !REG_EXPR(src2))
			return xor_none;
		if (!DECL_NAME(REG_EXPR(src2)))
			return xor_none;
		if (strncmp(DECL_NAME_POINTER(REG_EXPR(src2)), "rap_cookie", sizeof("rap_cookie") - 1))
			return xor_none;
	}

	return REGNO(dst) == REGNO(src1) ? ret : xor_none;
}

enum rap_mov { mov_none, mov_entry, mov_exit };

static enum rap_mov rap_match_mov(rtx body)
{
	rtx dst, src;
	tree expr;
	enum rap_mov ret;

	if (GET_CODE(body) != SET)
		return mov_none;

	dst = SET_DEST(body);
	if (!REG_P(dst) || !REG_EXPR(dst))
		return mov_none;
	expr = REG_EXPR(dst);
	gcc_assert(DECL_P(expr));
	if (!DECL_NAME(expr))
		return mov_none;
	if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_entry_", sizeof("rap_retaddr_entry_") - 1))
		ret = mov_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_exit_", sizeof("rap_retaddr_exit_") - 1))
		ret = mov_exit;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_entry", sizeof("rap_cookie_entry") - 1))
		ret = mov_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_exit", sizeof("rap_cookie_exit") - 1))
		ret = mov_exit;
	else
		return mov_none;

	src = SET_SRC(body);
	if (!REG_P(src) || !REG_EXPR(src))
		return mov_none;
	if (!DECL_NAME(REG_EXPR(src)) || !id_equal(DECL_NAME(REG_EXPR(src)), "rap_cookie"))
		return mov_none;
	gcc_assert(GET_MODE(dst) == GET_MODE(src));
	return ret;
}

static void rap_fuse_mov_xor(rtx_insn *mov1, rtx xor1)
{
	rtx body, src;

	body = PATTERN(xor1);
	if (GET_CODE(body) == PARALLEL)
		body = XVECEXP(body, 0, 0);
	src = SET_SRC(body);
	SET_SRC(PATTERN(mov1)) = XEXP(src, 1);
	df_insn_rescan(mov1);
}

static bool rap_xorreg_spills(rtx_insn *xor1, rtx xorreg1, rtx_insn *xor2, unsigned int xorregno2)
{
	df_ref use;

	for (use = DF_REG_USE_CHAIN(REGNO(xorreg1)); use; use = DF_REF_NEXT_REG(use)) {
		rtx_insn *insn;
		rtx body;

		if (DF_REF_IS_ARTIFICIAL(use))
			continue;

		insn = DF_REF_INSN(use);

		if (insn == xor1)
			continue;

		if (insn == xor2)
			continue;

		if (!insn_dominates_insn_p(CDI_DOMINATORS, xor1, insn))
			continue;
		if (!insn_dominates_insn_p(CDI_POST_DOMINATORS, xor2, insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);
		if (GET_CODE(body) != SET)
			continue;
		if (!REG_P(SET_SRC(body)))
			continue;
		if (ORIGINAL_REGNO(xorreg1) != ORIGINAL_REGNO(SET_SRC(body)))
			continue;

		switch (GET_CODE(SET_DEST(body))) {
		default:
			print_rtl_single(stderr, insn);
			gcc_unreachable();

		case MEM:
			return true;

		case REG:
			if (REGNO(SET_DEST(body)) == REGNO(SET_SRC(body))) {
#if BUILDING_GCC_VERSION >= 4008
				continue;
#else
				gcc_unreachable();
#endif
			}
			if (ORIGINAL_REGNO(SET_DEST(body)) == ORIGINAL_REGNO(SET_SRC(body)))
#if BUILDING_GCC_VERSION >= 4008
				continue;
#else
				gcc_unreachable();
#endif

			if (rap_xorreg_spills(insn, SET_DEST(body), xor2, xorregno2))
				return true;
			continue;
		}
	}

//	gcc_assert(ORIGINAL_REGNO(xorreg1) == xorregno2);
	return false;
}

// eliminate the two xors if the result register can't be clobbered in-between
static unsigned int rap_xor_elimination_execute(void)
{
	unsigned int i, n_rap_xors;
	struct rap_xor {
		basic_block movbb;
		rtx_insn *movinsn;
		rtx movreg;

		basic_block xorbb;
		rtx_insn *xorinsn;
		rtx xorreg;
	} *rap_xors;
	df_ref use;
	rtx_insn *barrier = 0;

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	if (!optimize) {
		df_finish_pass(true);
		df_scan_add_problem();
		df_scan_alloc(NULL);

		df_lr_add_problem();
		df_live_add_problem();
		df_live_set_all_dirty();

		df_chain_add_problem(DF_DU_CHAIN | DF_UD_CHAIN);
		df_md_add_problem();
#if BUILDING_GCC_VERSION >= 6000
		df_mir_add_problem();
#endif
		df_note_add_problem();

		df_rd_add_problem();
		df_scan_blocks();
		df_analyze();
	}

//	n_rap_xors = 1 + MIN(rap_xor_checks[0] + rap_xor_checks[1], EDGE_COUNT(EXIT_BLOCK_PTR_FOR_FN(cfun)->preds));
	n_rap_xors = 1 + rap_xor_checks[0] + rap_xor_checks[1];
	rap_xors = XCNEWVEC(struct rap_xor, n_rap_xors);

	for (use = DF_REG_USE_CHAIN(rap_reg_id); use; use = DF_REF_NEXT_REG(use)) {
		rtx_insn *insn;
		rtx body;

		if (DF_REF_IS_ARTIFICIAL(use))
			continue;

		insn = DF_REF_INSN(use);
		if (!NONJUMP_INSN_P(insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);

		if (GET_CODE(body) != SET)
			continue;
		if (GET_CODE(SET_SRC(body)) != ASM_OPERANDS)
			continue;
		if (!REG_P(SET_DEST(body)))
			continue;
		if (REGNO(SET_DEST(body)) != rap_reg_id)
			continue;

		gcc_assert(!barrier);
		barrier = insn;
		break;
	}

	for (i = 0, use = DF_REG_USE_CHAIN(rap_reg_id); use; use = DF_REF_NEXT_REG(use)) {
		rtx_insn *insn;
		rtx body;

		if (DF_REF_IS_ARTIFICIAL(use))
			continue;

		insn = DF_REF_INSN(use);
		if (!NONJUMP_INSN_P(insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);

		if (GET_CODE(body) != SET)
			continue;
		if (GET_CODE(SET_SRC(body)) == ASM_OPERANDS)
			continue;

		if (rap_match_xor(body, true) != xor_none) {
			rap_xors[i].xorbb = DF_REF_BB(use);
			rap_xors[i].xorinsn = insn;
			rap_xors[i].xorreg = SET_DEST(body);
		} else if (rap_match_mov(body) != mov_none) {
			df_ref use2;

			rap_xors[i].movbb = DF_REF_BB(use);
			rap_xors[i].movinsn = insn;
			rap_xors[i].movreg = SET_DEST(body);

			for (use2 = DF_REG_USE_CHAIN(REGNO(rap_xors[i].movreg)); use2; use2 = DF_REF_NEXT_REG(use2)) {
				rtx_insn *insn;
				rtx body;

				if (DF_REF_IS_ARTIFICIAL(use2))
					continue;

				insn = DF_REF_INSN(use2);
				if (!NONJUMP_INSN_P(insn))
					continue;

				if (rap_xors[i].movbb != DF_REF_BB(use2))
					continue;
				gcc_assert(barrier);
				if (insn_dominates_insn_p(CDI_POST_DOMINATORS, rap_xors[i].movinsn, barrier) !=
				    insn_dominates_insn_p(CDI_POST_DOMINATORS, insn, barrier))
					continue;

				body = PATTERN(insn);
				if (GET_CODE(body) == PARALLEL)
					body = XVECEXP(body, 0, 0);

				if (rap_match_xor(body, false) != xor_none) {
					rap_xors[i].xorbb = DF_REF_BB(use2);
					rap_xors[i].xorinsn = insn;
					rap_xors[i].xorreg = SET_DEST(body);
					break;
				}
			}
			gcc_assert(rap_xors[i].xorbb);
			gcc_assert(rap_xors[i].xorinsn);
		} else {
			debug_rtx(insn);
			gcc_unreachable();
		}

		// keep the single dominating entry instrumentation at the beginning
		if (i && insn_dominates_insn_p(CDI_DOMINATORS, rap_xors[0].xorinsn, rap_xors[i].xorinsn)) {
			struct rap_xor temp = rap_xors[0];
			rap_xors[0] = rap_xors[i];
			rap_xors[i] = temp;
		}

		i++;
	}

	if (!i)
		goto out;
	gcc_assert(barrier);

	if (i == n_rap_xors - rap_xor_checks[0])
		n_rap_xors -= rap_xor_checks[0];

	if (i != n_rap_xors) {
		while (i--)
			debug_rtx(rap_xors[i].xorinsn);
		gcc_unreachable();
	}

	for (i = 1; i < n_rap_xors; i++)
		if (rap_xorreg_spills(rap_xors[0].xorinsn, rap_xors[0].xorreg, rap_xors[i].xorinsn, ORIGINAL_REGNO(rap_xors[i].xorreg)))
			goto out;

	for (i = 0; i < n_rap_xors; i++) {
		if (rap_xors[i].movinsn)
			rap_fuse_mov_xor(rap_xors[i].movinsn, rap_xors[i].xorinsn);
		delete_insn_and_edges(rap_xors[i].xorinsn);
	}

out:
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	free(rap_xors);
	return 0;
}

#define PASS_NAME rap_xor_elimination
#define TODO_FLAGS_FINISH TODO_dump_func | TODO_verify_rtl_sharing | TODO_verify_flow | TODO_df_verify | TODO_df_finish
#include "gcc-generate-rtl-pass.h"
#endif
