/*
 * Copyright 2011-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to help implement various PaX features
 *
 * - track lowest stack pointer
 *
 * TODO:
 * - move instrumentation to a single RTL pass and also inline it
 *
 * BUGS:
 * - none known
 */

#include "gcc-common.h"
#include "gcc-ksyms.h"

__visible int plugin_is_GPL_compatible;

#define track_frame_size (2 * tree_to_uhwi(SMP_CACHE_BYTES) - sizeof(HOST_WIDE_INT))
static const char track_function[] = "pax_track_stack";
#if defined(TARGET_386) || defined(AARCH64_ABI_DEFAULT)
static char *asm_call_track_function;
#endif
static GTY(()) tree track_function_decl;
static GTY(()) tree current_stack_pointer_decl;

static struct plugin_info stackleak_plugin_info = {
	.version	= "202208260920",
	.help		= "stackleak plugin\n"
};

static void stackleak_add_instrumentation(gimple_stmt_iterator *gsi)
{
	gimple_stmt_ptr stmt;

#if defined(TARGET_386) || defined(AARCH64_ABI_DEFAULT)
	location_t loc;
	tree input1, input2, fnref, output;
#if BUILDING_GCC_VERSION <= 4007
	VEC(tree, gc) *inputs = NULL;
	VEC(tree, gc) *outputs = NULL;
#else
	vec<tree, va_gc> *inputs = NULL;
	vec<tree, va_gc> *outputs = NULL;
#endif

	gcc_assert(current_stack_pointer_decl);
	add_referenced_var(current_stack_pointer_decl);

	// asm volatile("pax_direct_call %c1" : "=r"(current_stack_pointer) : "i"(track_function), "0"(current_stack_pointer));
	input1 = build_tree_list(NULL_TREE, build_const_char_string(2, "i"));
	if (gsi_end_p(*gsi))
		loc = DECL_SOURCE_LOCATION(current_function_decl);
	else
		loc = gimple_location(gsi_stmt(*gsi));
	fnref = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(track_function_decl)), track_function_decl);
	input1 = chainon(NULL_TREE, build_tree_list(input1, fnref));

	input2 = build_tree_list(NULL_TREE, build_const_char_string(2, "0"));
	input2 = chainon(NULL_TREE, build_tree_list(input2, current_stack_pointer_decl));

	output = build_tree_list(NULL_TREE, build_const_char_string(3, "=r"));
	output = chainon(NULL_TREE, build_tree_list(output, current_stack_pointer_decl));

	VEC_safe_push(tree, gc, inputs, input1);
	VEC_safe_push(tree, gc, inputs, input2);
	VEC_safe_push(tree, gc, outputs, output);

	if (!asm_call_track_function)
		gcc_assert(asprintf(&asm_call_track_function, "%spax_direct_call %%c1\n\t", RAP_SAFE_ASM ? TREE_STRING_POINTER(RAP_SAFE_ASM) : "") != -1);
	stmt = gimple_build_asm_vec(asm_call_track_function, inputs, outputs, NULL, NULL);

	gimple_set_location(stmt, loc);
	gimple_asm_set_volatile(as_a_gasm(stmt), true);
#else
	// insert call to void pax_track_stack(void)
	stmt = gimple_build_call(track_function_decl, 0);
#endif

	gsi_insert_before(gsi, stmt, GSI_SAME_STMT);
}

static unsigned int stackleak_tree_instrument_execute(void)
{
	basic_block bb, entry_bb;
	bool is_leaf = true;
	gimple_stmt_iterator gsi;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	// 1. loop through BBs and GIMPLE statements
	if (warn_vla > 0)
		FOR_EACH_BB_FN(bb, cfun) {
			for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
				gimple_stmt_ptr stmt;

				stmt = gsi_stmt(gsi);
				if (!is_gimple_call(stmt))
					continue;

				is_leaf = false;

				// gimple match: align 8 built-in BUILT_IN_NORMAL:BUILT_IN_ALLOCA attributes <tree_list 0xb7576450>
				if (!gimple_alloca_call_p(stmt))
					continue;

				// 2. linux 4.20+ no longer permits alloca, bail out
				error_at(gimple_expansion_point_location(gsi_stmt(gsi)), "alloca is no longer permitted in linux");
			}
		}

	// special cases for some bad linux code: taking the address of static inline functions will materialize them
	// but we mustn't instrument some of them as the resulting stack alignment required by the function call ABI
	// will break other assumptions regarding the expected (but not otherwise enforced) register clobbering ABI.
	// case in point: native_save_fl on amd64 when optimized for size clobbers rdx if it were instrumented here.
	if (is_leaf && !TREE_PUBLIC(current_function_decl) && DECL_DECLARED_INLINE_P(current_function_decl))
		goto out;
	if (is_leaf && !strncmp(IDENTIFIER_POINTER(DECL_NAME(current_function_decl)), "_paravirt_", 10))
		goto out;

	// 3. insert track call at the beginning
	entry_bb = get_single_pred_entry_block();
	gsi = gsi_start_nondebug_after_labels_bb(entry_bb);
	stackleak_add_instrumentation(&gsi);

out:
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

static unsigned int stackleak_final_execute(void)
{
	rtx_insn *insn, *next;

	// 1. find pax_track_stack calls
	for (insn = get_insns(); insn; insn = next) {
		rtx body;
#if defined(TARGET_386) || defined(AARCH64_ABI_DEFAULT)
		rtx dest, src, src1, src2;
#endif

		next = NEXT_INSN(insn);

#if defined(TARGET_386) || defined(AARCH64_ABI_DEFAULT)
		// rtl match: (insn:TI 6 2 7 2 (parallel [ (asm_operands/v ("pax_direct_call %c1") ("") 0 [ (symbol_ref:DI ("pax_track_stack") [flags 0x41] <function_decl 0x7f2548aeed00 pax_track_stack>) ] [ (asm_input:DI ("i")) ] []) (clobber (reg:CCFP 18 fpsr)) (clobber (reg:CC 17 flags)) ]) -1 (expr_list:REG_UNUSED (reg:CCFP 18 fpsr) (expr_list:REG_UNUSED (reg:CC 17 flags) (nil))))
		if (GET_CODE(insn) != INSN)
			continue;
		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);
		if (GET_CODE(body) != SET)
			continue;

		dest = SET_DEST(body);
		if (!REG_P(dest))
			continue;
		if (REG_EXPR(dest) != current_stack_pointer_decl)
			continue;

		src = SET_SRC(body);
		if (GET_CODE(src) != ASM_OPERANDS)
			continue;
		if (!MEM_VOLATILE_P(src))
			continue;
		if (!asm_call_track_function)
			continue;
		if (strcmp(ASM_OPERANDS_TEMPLATE(src), asm_call_track_function))
			continue;
		if (ASM_OPERANDS_INPUT_LENGTH(src) != 2)
			continue;

		src1 = ASM_OPERANDS_INPUT(src, 0);
		if (GET_CODE(src1) != SYMBOL_REF)
			continue;
		gcc_assert(SYMBOL_REF_FUNCTION_P(src1));
		if (SYMBOL_REF_DECL(src1) != track_function_decl)
			continue;

		src2 = ASM_OPERANDS_INPUT(src, 1);
		if (!REG_P(src2))
			continue;
		if (REG_EXPR(src2) != current_stack_pointer_decl)
			continue;

//		warning(0, "track_frame_size: %d %ld %d", cfun->calls_alloca, get_frame_size(), track_frame_size);
		// 2. delete call
		delete_insn_and_edges(insn);
#else
		// 1. find pax_track_stack calls
		// rtl match: (call_insn 8 7 9 3 (call (mem (symbol_ref ("pax_track_stack") [flags 0x41] <function_decl 0xb7470e80 pax_track_stack>) [0 S1 A8]) (4)) -1 (nil) (nil))
		if (!CALL_P(insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);
		if (GET_CODE(body) != CALL)
			continue;

		body = XEXP(body, 0);
		if (GET_CODE(body) != MEM)
			continue;

		body = XEXP(body, 0);
		if (GET_CODE(body) != SYMBOL_REF)
			continue;
		gcc_assert(SYMBOL_REF_FUNCTION_P(body));
		if (SYMBOL_REF_DECL(body) != track_function_decl)
			continue;

		// 2. delete call
		remove_call_arg_locations(insn);
		delete_insn_and_edges(insn);
#endif
	}

//	print_simple_rtl(stderr, get_insns());
//	print_rtl(stderr, get_insns());
//	fflush(stderr);
//	warning(0, "track_frame_size: %d %ld %d", cfun->calls_alloca, get_frame_size(), track_frame_size);

	return 0;
}

static bool stackleak_track_stack_gate(void)
{
#ifdef TARGET_386
	if (ix86_cmodel != CM_KERNEL)
		return false;
#endif

	if (lookup_attribute("no_instrument_function", DECL_ATTRIBUTES(current_function_decl)))
		return false;

	if (has_trait(current_function_decl, "boot"))
		return false;

	return true;
}

static void stackleak_find_decls(void *event_data, void *data __unused)
{
	tree decl = (tree)event_data;

	if (decl == error_mark_node)
		return;

	if (!VAR_P(decl))
		return;
	gcc_assert(DECL_NAME_POINTER(decl));
	if (strcmp(DECL_NAME_POINTER(decl), "current_stack_pointer"))
		return;

	current_stack_pointer_decl = decl;
}

#if BUILDING_GCC_VERSION < 4007
static void stackleak_find_all_decls(void *event_data, void *data __unused)
{
	varpool_node_ptr vnode;

	FOR_EACH_VARIABLE(vnode)
		stackleak_find_decls(NODE_DECL(vnode), NULL);
}
#endif

static void stackleak_start_unit(void *gcc_data __unused, void *user_data __unused)
{
	tree fntype;

	// void pax_track_stack(void)
	fntype = build_function_type_list(void_type_node, NULL_TREE);
	build_pointer_type(fntype);
	track_function_decl = build_fn_decl(track_function, fntype);
	DECL_ASSEMBLER_NAME(track_function_decl); // for LTO
	TREE_PUBLIC(track_function_decl) = 1;
	TREE_USED(track_function_decl) = 1;
	DECL_EXTERNAL(track_function_decl) = 1;
	// Even though this is a compiler generated function declaration, we need
	// to mark it "normal" to get the RAP plugin to generate the required RAP
	// hashes. Otherwise building modules won't work because of the missing
	// __rap_hash_ret_pax_track_stack symbol only defined in vmlinux.
	DECL_ARTIFICIAL(track_function_decl) = 0;
	DECL_PRESERVE_P(track_function_decl) = 1;
	invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, track_function_decl);
}

static bool stackleak_tree_instrument_gate(void)
{
	return stackleak_track_stack_gate();
}

#define PASS_NAME stackleak_tree_instrument
#define PROPERTIES_REQUIRED PROP_gimple_leh | PROP_cfg
#define TODO_FLAGS_START TODO_verify_ssa | TODO_verify_flow | TODO_verify_stmts
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_update_ssa | TODO_rebuild_cgraph_edges
#include "gcc-generate-gimple-pass.h"

static poly_int64 get_total_frame_size(void)
{
	poly_int64 total_size;

#ifdef TARGET_386
	INITIAL_ELIMINATION_OFFSET(ARG_POINTER_REGNUM, STACK_POINTER_REGNUM, total_size);
#elif defined(AARCH64_ABI_DEFAULT)
	INITIAL_ELIMINATION_OFFSET(ARG_POINTER_REGNUM, STACK_POINTER_REGNUM, total_size);
#elif defined(TARGET_POWERPC64)
//	total_size = rs6000_stack_info()->total_size;
	INITIAL_ELIMINATION_OFFSET(ARG_POINTER_REGNUM, HARD_FRAME_POINTER_REGNUM, total_size);
#else
	total_size = get_frame_size();
#endif

	return total_size;
}

static bool stackleak_final_gate(void)
{
#if BUILDING_GCC_VERSION >= 8000
	HOST_WIDE_INT frame_size;
#endif
	HOST_WIDE_INT total_size;

	// linux 4.20+ no longer permits alloca, bail out
#if BUILDING_GCC_VERSION < 7000
	if (cfun->calls_alloca)
		return false;
#else
	gcc_assert(warn_vla <= 0 || !cfun->calls_alloca);
#endif

#if BUILDING_GCC_VERSION < 8000
	total_size = get_total_frame_size();
#else
	gcc_assert(get_total_frame_size().is_constant(&total_size));
	gcc_assert(get_frame_size().is_constant(&frame_size));
#endif

	// nothing to remove if the function wasn't instrumented to begin with
	if (!stackleak_track_stack_gate()) {
		return false;
	}

	// keep calls only if function frame is big enough
	if (total_size >= track_frame_size) {
		return false;
	}

	return true;
}

#define PASS_NAME stackleak_final
#define TODO_FLAGS_FINISH TODO_dump_func
#include "gcc-generate-rtl-pass.h"

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;

	static const struct ggc_root_tab gt_ggc_r_gt_stackleak[] = {
		INIT_GGC_ROOT_TAB_TREE(track_function_decl),
		INIT_GGC_ROOT_TAB_TREE(current_stack_pointer_decl),
		LAST_GGC_ROOT_TAB
	};

//	PASS_INFO(stackleak_tree_instrument, "tree_profile", 1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(stackleak_tree_instrument, "optimized", 1, PASS_POS_INSERT_BEFORE);
//	PASS_INFO(stackleak_final, "*free_cfg", 1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(stackleak_final, "final", 1, PASS_POS_INSERT_BEFORE);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &stackleak_plugin_info);

	register_callback(plugin_name, PLUGIN_START_UNIT, stackleak_start_unit, NULL);
#if BUILDING_GCC_VERSION < 4007
	register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, stackleak_find_all_decls, NULL);
#else
	register_callback(plugin_name, PLUGIN_FINISH_DECL, stackleak_find_decls, NULL);
#endif
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_stackleak);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &stackleak_tree_instrument_pass_info);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &stackleak_final_pass_info);

	return 0;
}
