/*
 * Copyright 2021-2022 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * eliminate a pointless variable and its copying that arise in code constructs
 * where a variable is initialized with a statement expression that takes its address
 * e.g., DECLARE_WAIT_QUEUE_HEAD_ONSTACK under LOCKDEP produces them
 */

#include "utilities.h"

static tree uses_deadvar(tree *tp, int *walk_subtrees, void *data)
{
	tree op;
	bitmap deadvars;
	struct walk_stmt_info *wi = (struct walk_stmt_info *)data;

	op = *tp;
	deadvars = (bitmap)wi->info;

	gcc_assert(!TYPE_P(op));
	if (!VAR_P(op))
		return NULL_TREE;

	*walk_subtrees = 0;
	if (!bitmap_bit_p(deadvars, DECL_UID(op)))
		return NULL_TREE;

	bitmap_clear_bit(deadvars, DECL_UID(op));
	return op;
}

unsigned int dead_copy_elimination_execute(void)
{
	basic_block bb;
	bitmap deadvars;

	deadvars = BITMAP_ALLOC(NULL);
	bitmap_clear(deadvars);

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;

			stmt = gsi_stmt(gsi);
			gimple_set_plf(stmt, GF_PLF_1, false);
			gimple_set_plf(stmt, GF_PLF_2, false);
		}
	}

	// 1. walk stmts to collect candidate stmts and variables
	//    mark stmts with a flag a and record the variable UIDs in a bitmap
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt, stmt2;
			gimple_stmt_iterator gsi2;
			tree deadvar, var;

			stmt = gsi_stmt(gsi);
			if (!gimple_assign_single_p(stmt))
				continue;

			deadvar = gimple_assign_lhs(stmt);

			if (TREE_STATIC(deadvar))
				continue;
			if (TREE_ADDRESSABLE(deadvar))
				continue;
			if (!VAR_P(deadvar))
				continue;
			if (!DECL_IGNORED_P(deadvar))
				continue;

			var = gimple_assign_rhs1(stmt);

			gsi2 = gsi;
			gsi_next(&gsi2);
			if (gsi_end_p(gsi2))
				continue;
			stmt2 = gsi_stmt(gsi2);
			if (!stmt2)
				continue;
			if (!gimple_assign_single_p(stmt2))
				continue;
			if (gimple_assign_lhs(stmt2) != var)
				continue;
			if (gimple_assign_rhs1(stmt2) != deadvar)
				continue;

			gimple_set_plf(stmt, GF_PLF_1, true);
			gimple_set_plf(stmt2, GF_PLF_2, true);
			bitmap_set_bit(deadvars, DECL_UID(deadvar));
		}
	}

	if (bitmap_empty_p(deadvars))
		goto cleanup;

	// 2. walk non-candidate stmts and look for marked variable uses
	//    unmark variable on any such uses
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			struct walk_stmt_info wi = {};
			gimple_stmt_ptr stmt;

			stmt = gsi_stmt(gsi);
			if (gimple_plf(stmt, GF_PLF_1))
				continue;
			if (gimple_plf(stmt, GF_PLF_2))
				continue;

			wi.info = deadvars;
			walk_gimple_stmt(&gsi, NULL, uses_deadvar, &wi);
		}
	}

	// 3. walk candidate stmts and remove them only if their variable is still marked
	//    log leftover deadvars for inspection?
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi);) {
			gimple_stmt_ptr stmt, stmt2;
			tree deadvar;

			stmt = gsi_stmt(gsi);
			if (!gimple_plf(stmt, GF_PLF_1))
				goto next;

			deadvar = gimple_assign_lhs(stmt);
			if (!bitmap_bit_p(deadvars, DECL_UID(deadvar)))
				goto next;

			gsi_next(&gsi);
			stmt2 = gsi_stmt(gsi);
			gcc_assert(stmt2 && gimple_plf(stmt2, GF_PLF_2));

			unlink_stmt_vdef(stmt2);
			gsi_remove(&gsi, true);
			gsi_prev(&gsi);
			unlink_stmt_vdef(stmt);
			gsi_remove(&gsi, true);
			continue;

next:
			gsi_next(&gsi);
		}
	}

cleanup:
	BITMAP_FREE(deadvars);
	return 0;
}

#define PASS_NAME dead_copy_elimination
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"
