/*
 * Copyright 2023 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * improved version of the 'error' and 'warning' attributes
 */

#include "utilities.h"

static void split_diagnose_attribute(tree args, tree *kind, tree *msg, tree *varargs)
{
	*kind = TREE_VALUE(args);

	args = TREE_CHAIN(args);
	*msg = TREE_VALUE(args);
	*varargs = TREE_CHAIN(args);
}

static tree handle_diagnose_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree fndecl, kind, msg, varargs;
	unsigned HOST_WIDE_INT arg_count;

	*no_add_attrs = true;
	fndecl = *node;

	if (TREE_CODE(fndecl) != FUNCTION_DECL) {
		error_at(DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION, "%qE attribute applies to functions only (%qD)", name, *node);
		return NULL_TREE;
	}

	split_diagnose_attribute(args, &kind, &msg, &varargs);

	if (TREE_CODE(kind) != STRING_CST) {
		error_at(DECL_SOURCE_LOCATION(fndecl), "%qE attribute needs a string for its first argument (%qE)", name, kind);
		return NULL_TREE;
	}
	if (strcmp(TREE_STRING_POINTER(kind), "error") &&
	    strcmp(TREE_STRING_POINTER(kind), "warning")) {
		error_at(DECL_SOURCE_LOCATION(fndecl), "%qE attribute needs 'error' or 'warning' for its first argument (%qE)", name, kind);
		return NULL_TREE;
	}

	if (TREE_CODE(msg) != STRING_CST) {
		error_at(DECL_SOURCE_LOCATION(fndecl), "%qE attribute needs a format string for its second argument (%qE)", name, msg);
		return NULL_TREE;
	}

	if (varargs == NULL_TREE) {
		error_at(DECL_SOURCE_LOCATION(fndecl), "%qE attribute has a second arugment with no further arguments (%E)", name, msg);
		return NULL_TREE;
	}

	if (list_length(varargs) > 2) {
		error_at(DECL_SOURCE_LOCATION(fndecl), "%qE attribute with a second arugment cannot handle more than 2 further arguments (%E)", name, msg);
		return NULL_TREE;
	}

	arg_count = type_num_arguments(TREE_TYPE(fndecl));
	for (; varargs; varargs = TREE_CHAIN(varargs)) {
		unsigned HOST_WIDE_INT cur_val;
		tree position = TREE_VALUE(varargs);

		if (TREE_CODE(position) != INTEGER_CST) {
			error_at(DECL_SOURCE_LOCATION(fndecl), "parameter %qE isn't an integer", position);
			return NULL_TREE;
		}

		cur_val = tree_to_uhwi(position);
		if (arg_count < cur_val) {
			error_at(DECL_SOURCE_LOCATION(fndecl), "parameter %qE is outside range.", position);
			return NULL_TREE;
		}
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec diagnose_attr = ATTRIBUTE_SPEC_INIT("diagnose", 2, -1, true, false, false, handle_diagnose_attribute, false, NULL);

unsigned int diagnose_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;
			tree fndecl, attr, kind, msg, varargs, arg1, arg2;
			char *format = NULL;
			unsigned int nargs;

			stmt = gsi_stmt(gsi);
			if (!is_gimple_call(stmt))
				continue;

			fndecl = gimple_call_fndecl(stmt);
			if (fndecl == NULL_TREE)
				continue;

			attr = lookup_attribute("diagnose", DECL_ATTRIBUTES(fndecl));
			if (attr == NULL_TREE)
				continue;

			attr = TREE_VALUE(attr);
			split_diagnose_attribute(attr, &kind, &msg, &varargs);
			gcc_assert(asprintf(&format, "call to '%s' diagnosed error: %s",
				lang_hooks.decl_printable_name(fndecl, 1),
				TREE_STRING_POINTER(msg)) != -1);

			nargs = list_length(varargs);
			switch (nargs) {
			default:
				debug_gimple_stmt(stmt);
				gcc_unreachable();

			case 1:
				arg1 = gimple_call_arg(stmt, TREE_INT_CST_LOW(TREE_VALUE(varargs)) - 1);
				error_at(gimple_inline_point_location(stmt), format, arg1);
				break;

			case 2:
				arg1 = gimple_call_arg(stmt, TREE_INT_CST_LOW(TREE_VALUE(varargs)) - 1);
				arg2 = gimple_call_arg(stmt, TREE_INT_CST_LOW(TREE_VALUE(TREE_CHAIN(varargs))) - 1);
				error_at(gimple_inline_point_location(stmt), format, arg1, arg2);
				break;
			}

			free(format);
		}
	}

	return 0;
}

#define PASS_NAME diagnose
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_dump_func
#include "gcc-generate-gimple-pass.h"
