/*
 * Copyright 2020-2022 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * move marked automatic (local) variables off the stack
 *
 * a mark is either a type attribute where all instances of the type must be moved,
 * or a variable attribute for a specific variable to move,
 * or a function attribute marking pointer arguments whose targets must be moved
 *
 * an IPA pass helps propagate the function arg attribute to actual variables,
 * the rest needs manual annotation
 *
 * use case: private (unreadable) kernel stacks don't allow access from foreign threads
 *           so wait queue and other structures must not be allocated there
 *
 * guidance on how to use the 'nolocal' and 'nolocal_arg' attributes:
 *
 * - the obvious preference is automatic discovery of nolocal variables but static
 *   analysis can go only so far, hence the need to augment/guide the analysis.
 *
 * - the 'nolocal' attribute on a type is the big hammer approach as it eliminates any
 *   shortcomings of static analysis at the cost of converting all local variables of
 *   the type. therefore it should only be used when most (if not all) local variables
 *   of the type are accessed remotely. the textbook case is 'struct work_struct'.
 *
 *   a counterexample is 'struct list_head' which is the type of a field of 'struct work_struct'
 *   (and thus one may be tempted to mark it as 'nolocal') but is also used in many
 *   other places that need not be moved off the stack and thus would cause unnecessary
 *   nolocal transformations and performance impact.
 *
 * - the 'nolocal_arg' attribute is necessary to allow static analysis to trace back
 *   data flow from sinks (remote stack accesses) to sources (remotely accessed local
 *   variables). in an ideal world the need for 'nolocal_arg' is discovered by static
 *   analysis and thus explicit placement of 'nolocal_arg' should be elimininated as
 *   analysis improves.
 *
 * - the 'nolocal' attribute on a (local) variable is considered as a last resort where
 *   the variable's type itself cannot be marked 'nolocal' nor can 'nolocal_arg' and
 *   static analysis track back the data flow to the variable. in an ideal world this
 *   use of 'nolocal' is not necessary and improvements to static analysis can be
 *   measured by the reduction of 'nolocal' uses on variables.
 */

#include "utilities.h"
#include "gcc-ksyms.h"

struct nolocal_var {
	tree oldvar;
	tree newvar;
};

bool report_nolocals = false;

static GTY(()) tree alloc_nolocal_decl;
static GTY(()) tree free_nolocal_decl;

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_gt_nolocal[] = {
	INIT_GGC_ROOT_TAB_TREE(alloc_nolocal_decl),
	INIT_GGC_ROOT_TAB_TREE(free_nolocal_decl),
	LAST_GGC_ROOT_TAB
};

static bool is_nolocal_var(tree oldvar)
{
	tree type;

	if (!VAR_P(oldvar))
		return false;
	if (TREE_STATIC(oldvar))
		return false;
	if (DECL_EXTERNAL(oldvar))
		return false;
	if (!TREE_ADDRESSABLE(oldvar))
		return false;

	if (lookup_attribute("nolocal", DECL_ATTRIBUTES(oldvar)))
		return true;

	type = TREE_TYPE(oldvar);
	if (lookup_attribute("nolocal", TYPE_ATTRIBUTES(type)))
		return true;

	type = strip_array_types(type);
	if (lookup_attribute("nolocal", TYPE_ATTRIBUTES(type)))
		return true;

	return false;
}

static int (*old_nolocal_gimplify_expr) (tree *expr_p, gimple_seq *pre_p, gimple_seq *post_p);
static int nolocal_gimplify_expr(tree *expr_p, gimple_seq *pre_p, gimple_seq *post_p)
{
	tree stmt = *expr_p;

	if (TREE_CODE(stmt) == DECL_EXPR && is_nolocal_var(DECL_EXPR_DECL(stmt))) {
		tree decl = DECL_EXPR_DECL(stmt);
		tree id = get_identifier("uninitialized");
		tree attr = DECL_ATTRIBUTES(decl);

		DECL_ATTRIBUTES(decl) = tree_cons(id, NULL_TREE, attr);
	}
	return old_nolocal_gimplify_expr(expr_p, pre_p, post_p);
}

void nolocal_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	bitmap_obstack_initialize(NULL);

	old_nolocal_gimplify_expr = lang_hooks.gimplify_expr;
	lang_hooks.gimplify_expr = nolocal_gimplify_expr;
}

void nolocal_find_decls(void *event_data, void *data __unused)
{
	tree decl = (tree)event_data;

	if (seen_error())
		return;

	if (decl == NULL_TREE || decl == error_mark_node)
		return;

	switch (TREE_CODE(decl)) {
	default:
		return;

	case FUNCTION_DECL:
		if (DECL_ARTIFICIAL(decl))
			return;

		if (DECL_ABSTRACT_ORIGIN(decl) && DECL_ABSTRACT_ORIGIN(decl) != decl)
			return;

		gcc_assert(DECL_NAME_POINTER(decl)[0]);

		if (id_equal(DECL_NAME(decl), "alloc_nolocal")) {
			if (alloc_nolocal_decl == NULL_TREE)
				alloc_nolocal_decl = decl;
			else
				gcc_assert(alloc_nolocal_decl == decl);
			return;
		}

		if (id_equal(DECL_NAME(decl), "free_nolocal")) {
			if (free_nolocal_decl == NULL_TREE)
				free_nolocal_decl = decl;
			else
				gcc_assert(free_nolocal_decl == decl);
			return;
		}
	}
}

#if BUILDING_GCC_VERSION < 4007
void nolocal_find_all_decls(void *event_data, void *data __unused)
{
	cgraph_node_ptr cnode;

	FOR_EACH_FUNCTION(cnode) {
		tree decl;

		if (NODE_THUNK(cnode) || cnode->alias)
			continue;
		if (NODE_GLOBAL(cnode)->inlined_to)
			continue;

		for (decl = NODE_DECL(cnode); decl && !alloc_nolocal_decl; decl = TREE_CHAIN(decl))
			nolocal_find_decls(decl, NULL);
		for (decl = NODE_DECL(cnode); decl && !free_nolocal_decl; decl = TREE_CHAIN(decl))
			nolocal_find_decls(decl, NULL);
	}

	if (alloc_nolocal_decl)
		invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, alloc_nolocal_decl);
	if (free_nolocal_decl)
		invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, free_nolocal_decl);
}
#endif

static tree handle_nolocal_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	location_t loc;
	tree decl;
	bool add_attrs = true;

	loc = DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION;
	*no_add_attrs = true;

	if (TYPE_P(*node)) {
		if (args == NULL_TREE)
			*no_add_attrs = false;
		else
			error_at(loc, "%qE attribute on %qD must not have arguments", name, *node);
		return NULL_TREE;
	}

	decl = *node;
	switch (TREE_CODE(decl)){
	default:
		error_at(loc, "%qE attribute applies to types, local variables and function declarations only (%qE)", name, decl);
		debug_tree(decl);
		return NULL_TREE;

	case TYPE_DECL:
		if (TREE_CODE(TREE_TYPE(decl)) != FUNCTION_TYPE) {
			error_at(loc, "%qE attribute on %qD applies to function types only", name, decl);
			return NULL_TREE;
		}
		// FALLTHROUGH

	case FUNCTION_DECL: {
		tree fntype, fnarg;
		unsigned int argnum, nnolocals;
		sbitmap nolocals;
		function_args_iterator args_iter;

		if (args == NULL_TREE) {
			error_at(loc, "%qE attribute on %qD must have arguments", name, decl);
			return NULL_TREE;
		}

		fntype = TREE_TYPE(decl);
		nnolocals = list_length(TYPE_ARG_TYPES(fntype));
		if (tree_last(TYPE_ARG_TYPES(fntype)) != void_list_node)
			nnolocals++;
		nolocals = sbitmap_alloc(nnolocals);
		sbitmap_clear(nolocals);

		for (; args; args = TREE_CHAIN(args)) {
			tree id;

			id = TREE_VALUE(args);
			if (TREE_CODE(id) != INTEGER_CST) {
				add_attrs = false;
				error_at(loc, "%qE argument %qE must be an integer", name, id);
				continue;
			}

			// verify that the function's nolocal params exist
			argnum = TREE_INT_CST_LOW(id);
			if (argnum >= nnolocals) {
				add_attrs = false;
				error_at(loc, "%qE argument %qE is too large (%u)", name, id, nnolocals);
				continue;
			}

			sbitmap_set_bit(nolocals, argnum);
		}

		// verify that the function's nolocal params are of a ptr type
		argnum = 0;
		FOREACH_FUNCTION_ARGS(fntype, fnarg, args_iter) {
			argnum++;

			if (argnum >= nnolocals)
				continue;
			if (!sbitmap_bit_p(nolocals, argnum))
				continue;
			if (POINTER_TYPE_P(fnarg))
				continue;

			add_attrs = false;
			error_at(loc, "%qE nolocal argument %u must be a pointer (%qT)", name, argnum, fnarg);
		}

		sbitmap_free(nolocals);
		break;
	}

	case VAR_DECL:
		if (!TREE_STATIC(decl))
			break;

		error_at(loc, "%qE attribute applies to automatic variables only (%qE)", name, decl);
		return NULL_TREE;
	}

	*no_add_attrs = !add_attrs;
	return NULL_TREE;
}

struct attribute_spec nolocal_attr = ATTRIBUTE_SPEC_INIT("nolocal", 0, -1, false, false, false, handle_nolocal_attribute, false, NULL);

static void collect_types(bitmap types, tree type)
{
	tree field;

	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		tree fieldtype;

		fieldtype = strip_array_types(TREE_TYPE(field));
		if (!RECORD_OR_UNION_TYPE_P(fieldtype))
			continue;

		gcc_assert(TYPE_CANONICAL(fieldtype));
		bitmap_set_bit(types, TYPE_UID(TYPE_CANONICAL(fieldtype)));
		collect_types(types, fieldtype);
	}
}

static bool __is_nolocal_type(bitmap types, tree type)
{
	bool ret = false;
	tree field;

	if (type == error_mark_node)
		return false;
	if (lookup_attribute("nolocal", TYPE_ATTRIBUTES(type)))
		return true;

	if (!RECORD_OR_UNION_TYPE_P(type))
		return false;

	for (field = TYPE_FIELDS(type); !ret && field; field = TREE_CHAIN(field)) {
		tree arg, fieldtype;

		fieldtype = strip_array_types(TREE_TYPE(field));
		if (__is_nolocal_type(types, fieldtype)) {
			ret = true;
			break;
		}

		if (!POINTER_TYPE_P(fieldtype))
			continue;

		fieldtype = strip_array_types(TREE_TYPE(fieldtype));
		if (lookup_attribute("nolocal", TYPE_ATTRIBUTES(type))) {
			ret = true;
			break;
		}

		if (!FUNC_OR_METHOD_TYPE_P(fieldtype))
			continue;

		for (arg = TYPE_ARG_TYPES(fieldtype); arg != NULL_TREE; arg = TREE_CHAIN(arg)) {
			tree argtype;

			argtype = TREE_VALUE(arg);
			if (!POINTER_TYPE_P(argtype))
				continue;

			argtype = strip_array_types(TREE_TYPE(argtype));
			if (!lookup_attribute("nolocal", TYPE_ATTRIBUTES(argtype)))
				continue;

			gcc_assert(TYPE_CANONICAL(argtype));
			if (!bitmap_bit_p(types, TYPE_UID(TYPE_CANONICAL(argtype))))
				continue;

			ret = true;
			break;
		}
	}

	return ret;
}

static bool is_nolocal_type(tree type)
{
	bitmap types;
	bool ret;

	if (type == error_mark_node)
		return false;

	type = strip_array_types(type);
	if (lookup_attribute("nolocal", TYPE_ATTRIBUTES(type)))
		return true;

	if (!RECORD_OR_UNION_TYPE_P(type))
		return false;

	types = BITMAP_ALLOC(NULL);
	bitmap_clear(types);
	collect_types(types, type);

	ret = __is_nolocal_type(types, type);

	BITMAP_FREE(types);
	return ret;
}

void nolocal_finish_type(void *event_data, void *data __unused)
{
	tree type = (tree)event_data;

	if (type == NULL_TREE || type == error_mark_node)
		return;

#if BUILDING_GCC_VERSION >= 5000
	if (TREE_CODE(type) == ENUMERAL_TYPE)
		return;
#endif

	if (TYPE_FIELDS(type) == NULL_TREE)
		return;
	if (lookup_attribute("nolocal", TYPE_ATTRIBUTES(type)))
		return;

	if (!is_nolocal_type(type))
		return;

	add_type_attr(type, "nolocal", NULL_TREE);
}

static tree get_nolocal_attr(gimple_stmt_ptr stmt)
{
	tree fndecl, fntype, attr = NULL_TREE;

	fndecl = gimple_call_fndecl(stmt);
	if (fndecl != NULL_TREE)
		attr = lookup_attribute("nolocal", DECL_ATTRIBUTES(fndecl));
	if (attr != NULL_TREE)
		return attr;

	fntype = gimple_call_fntype(stmt);
	if (fntype == NULL_TREE)
		return NULL_TREE;

	attr = lookup_attribute("nolocal", TYPE_ATTRIBUTES(fntype));
	if (attr != NULL_TREE)
		return attr;

	fntype = TYPE_NAME(fntype);
	if (fntype && TREE_CODE(fntype) == TYPE_DECL)
		attr = lookup_attribute("nolocal", DECL_ATTRIBUTES(fntype));

	return attr;
}

static tree uses_nolocal_var(tree *tp, int *walk_subtrees, void *data)
{
	tree op, var;
	struct walk_stmt_info *wi = (struct walk_stmt_info *)data;

	op = *tp;
	var = (tree)wi->info;

	gcc_assert(!TYPE_P(op));
	if (!VAR_P(op) && TREE_CODE(op) != PARM_DECL)
		return NULL_TREE;

	*walk_subtrees = 0;
	return op == var ? op : NULL_TREE;
}

static tree convert_nolocal_var(tree *tp, int *walk_subtrees, void *data)
{
	tree op, oldvar, newvar;
	struct walk_stmt_info *wi = (struct walk_stmt_info *)data;
	struct nolocal_var *vars = (struct nolocal_var *)wi->info;
	gimple_stmt_ptr stmt = gsi_stmt(wi->gsi);

	gcc_assert(stmt);
	op = *tp;
	oldvar = vars->oldvar;
	newvar = vars->newvar;

	switch (TREE_CODE(op)) {
	default:
		debug_gimple_stmt(stmt);
		debug_tree(op);
		debug_tree(oldvar);
		gcc_unreachable();

	case BIT_FIELD_REF:
	case VIEW_CONVERT_EXPR:
		break;

	case EQ_EXPR:
	case GE_EXPR:
	case GT_EXPR:
	case LE_EXPR:
	case LT_EXPR:
	case NE_EXPR:
	case IMAGPART_EXPR:
	case REALPART_EXPR:
		*walk_subtrees = 0;
		gcc_assert(TREE_OPERAND(op, 0) != oldvar);
		break;

	case CASE_LABEL_EXPR:
	case COMPLEX_CST:
	case FIELD_DECL:
	case FUNCTION_DECL:
	case INTEGER_CST:
	case LABEL_DECL:
	case RESULT_DECL:
	case SSA_NAME:
	case STRING_CST:
		*walk_subtrees = 0;
		break;

#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
	case TARGET_MEM_REF: {
		tree op0;

		op0 = TREE_OPERAND(op, 0);
		gcc_assert(get_inner_ref(op0) == get_inner_arg(op0));
		break;
	}

	case ARRAY_REF: {
		tree base, offset;
		poly_int64_pod bitsize, bitpos;
		enum machine_mode mode;
		int unsignedp, reversep, volatilep;
		gimple_stmt_ptr stmt2;
		gimple_stmt_iterator gsi;

		base = get_inner_reference(op, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
		if (!VAR_P(base))
			break;

		*walk_subtrees = 0;
		if (base != oldvar)
			break;

		gcc_assert(is_gimple_assign(stmt));
		gcc_assert(gimple_assign_single_p(stmt));
		gcc_assert(multiple_p(bitpos, BITS_PER_UNIT));

		if (offset == NULL_TREE) {
			tree newvar_cast;

			build_pointer_type(TREE_TYPE(op));
			newvar_cast = create_tmp_var(TYPE_POINTER_TO(TREE_TYPE(op)), "nolocal_var_cast");
			mark_sym_for_renaming(newvar_cast);
			stmt2 = gimple_build_assign(newvar_cast, newvar);
			gimple_set_block(stmt2, gimple_block(stmt));
			gimple_set_location(stmt2, gimple_location(stmt));
			gsi = gsi_for_stmt(stmt);
			gsi_insert_before(&gsi, stmt2, GSI_SAME_STMT);

			*tp = build2_loc(gimple_location(stmt), MEM_REF, TREE_TYPE(op), newvar_cast, build_int_cst(TREE_TYPE(newvar_cast), exact_div(bitpos, BITS_PER_UNIT)));
			update_stmt(stmt);
		} else if (TREE_CODE(offset) == NOP_EXPR || TREE_CODE(offset) == MULT_EXPR || TREE_CODE(offset) == PLUS_EXPR) {
			tree newvar_offset, newvar_adjusted;

			gsi = gsi_for_stmt(stmt);
			newvar_offset = force_gimple_operand_gsi(&gsi, offset, true, NULL, true, GSI_SAME_STMT);
			gcc_assert(TREE_CODE(newvar_offset) == SSA_NAME);

			newvar_adjusted = create_tmp_var(TREE_TYPE(newvar), "nolocal_var_adjusted");
			mark_sym_for_renaming(newvar_adjusted);
			stmt2 = gimple_build_assign_with_ops(POINTER_PLUS_EXPR, newvar_adjusted, newvar, newvar_offset);
			gimple_set_block(stmt2, gimple_block(stmt));
			gimple_set_location(stmt2, gimple_location(stmt));
			gsi_insert_before(&gsi, stmt2, GSI_SAME_STMT);

			*tp = build2_loc(gimple_location(stmt), MEM_REF, TREE_TYPE(op), newvar_adjusted, build_int_cst(TREE_TYPE(newvar_adjusted), exact_div(bitpos, BITS_PER_UNIT)));
			update_stmt(stmt);
		} else {
			debug_gimple_stmt(stmt);
			debug_tree(op);
			debug_tree(base);
			debug_tree(offset);
			gcc_unreachable();
		}

		break;
	}

	case ADDR_EXPR: {
		tree op0;

		op0 = TREE_OPERAND(op, 0);
		switch (TREE_CODE(op0)) {
		default:
			debug_gimple_stmt(stmt);
			debug_tree(op);
			debug_tree(op0);
			gcc_unreachable();

		case FUNCTION_DECL:
		case LABEL_DECL:
		case STRING_CST:
			*walk_subtrees = 0;
			break;

		case ARRAY_REF:
		case COMPONENT_REF: {
			tree base;
			tree offset;
			poly_int64_pod bitsize, bitpos;
			enum machine_mode mode;
			int unsignedp, reversep, volatilep;

			base = get_inner_reference(op0, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
			if (!VAR_P(base))
				break;

			*walk_subtrees = 0;
			if (base != oldvar)
				break;

			gcc_assert(multiple_p(bitpos, BITS_PER_UNIT));

			switch (gimple_code(stmt)) {
			default:
				debug_gimple_stmt(stmt);
				debug_tree(oldvar);
				debug_tree(base);
				gcc_unreachable();

			case GIMPLE_ASM:
				gcc_assert(offset == NULL_TREE);
				if (known_eq_poly_int64(bitpos, 0))
					*tp = newvar;
				else
					*tp = build2_loc(gimple_location(stmt), POINTER_PLUS_EXPR, TREE_TYPE(newvar), newvar, size_int(exact_div(bitpos, BITS_PER_UNIT)));
				update_stmt(stmt);
				break;

			case GIMPLE_ASSIGN: {
				enum tree_code offset_code;
				gimple_stmt_iterator gsi;
				gimple_stmt_ptr stmt2;

				gcc_assert(gimple_assign_rhs1(stmt) == op);

				offset_code = offset == NULL_TREE ? MAX_TREE_CODES : TREE_CODE(offset);
				switch (offset_code) {
				default:
					debug_gimple_stmt(stmt);
					debug_tree(op);
					debug_tree(base);
					debug_tree(offset);
					gcc_unreachable();

				case MAX_TREE_CODES:
					if (known_eq_poly_int64(bitpos, 0)) {
						*tp = newvar;
						gimple_assign_set_rhs_code(stmt, TREE_CODE(oldvar));
						update_stmt(stmt);
					} else {
						tree lhs;

						lhs = gimple_assign_lhs(stmt);
						stmt2 = gimple_build_assign_with_ops(POINTER_PLUS_EXPR, lhs, newvar, size_int(exact_div(bitpos, BITS_PER_UNIT)));
						gimple_set_block(stmt2, gimple_block(stmt));
						gimple_set_location(stmt2, gimple_location(stmt));
						gsi = gsi_for_stmt(stmt);
						gsi_replace(&gsi, stmt2, false);
					}
					break;

				case MULT_EXPR:
				case NOP_EXPR:
				case PLUS_EXPR:
				case SSA_NAME: {
					tree newvar_offset, newvar_adjusted, lhs;

					gsi = gsi_for_stmt(stmt);
					newvar_offset = force_gimple_operand_gsi(&gsi, offset, true, NULL, true, GSI_SAME_STMT);
					gcc_assert(TREE_CODE(newvar_offset) == SSA_NAME);

					newvar_adjusted = create_tmp_var(TREE_TYPE(newvar), "nolocal_var_adjusted");
					mark_sym_for_renaming(newvar_adjusted);
					stmt2 = gimple_build_assign_with_ops(POINTER_PLUS_EXPR, newvar_adjusted, newvar, newvar_offset);
					gimple_set_block(stmt2, gimple_block(stmt));
					gimple_set_location(stmt2, gimple_location(stmt));
					gsi_insert_before(&gsi, stmt2, GSI_SAME_STMT);

					lhs = gimple_assign_lhs(stmt);
					if (known_eq_poly_int64(bitpos, 0))
						stmt2 = gimple_build_assign(lhs, newvar_adjusted);
					else
						stmt2 = gimple_build_assign_with_ops(POINTER_PLUS_EXPR, lhs, newvar_adjusted, size_int(exact_div(bitpos, BITS_PER_UNIT)));
					gimple_set_block(stmt2, gimple_block(stmt));
					gimple_set_location(stmt2, gimple_location(stmt));
					gsi = gsi_for_stmt(stmt);
					gsi_replace(&gsi, stmt2, false);

					break;
				}
				}
				break;
			}

			case GIMPLE_COND:
				fputs("COMPONENT ", stderr);
				print_dec(bitpos, stderr);
				fputs(" ", stderr);
				debug_gimple_stmt(stmt);
				debug_tree(op);
				debug_tree(op0);
				debug_tree(base);
				debug_tree(offset);
				gcc_unreachable();
				break;
			}

			break;
		}

		case MEM_REF: {
			tree base, offset;
			poly_int64_pod bitsize, bitpos;
			enum machine_mode mode;
			int unsignedp, reversep, volatilep;

			base = get_inner_reference(op0, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
			if (!VAR_P(base))
				break;

			*walk_subtrees = 0;
			if (base != oldvar)
				break;

			gcc_assert(is_gimple_assign(stmt));
			gcc_assert(multiple_p(bitpos, BITS_PER_UNIT));
			gcc_assert(offset == NULL_TREE);

			if (known_eq_poly_int64(bitpos, 0)) {
				*tp = newvar;
				gimple_assign_set_rhs_code(stmt, TREE_CODE(oldvar));
				update_stmt(stmt);
			} else {
				gimple_stmt_ptr stmt2;
				tree lhs;

				lhs = gimple_assign_lhs(stmt);
				stmt2 = gimple_build_assign_with_ops(POINTER_PLUS_EXPR, lhs, newvar, size_int(exact_div(bitpos, BITS_PER_UNIT)));
				gimple_set_block(stmt2, gimple_block(stmt));
				gimple_set_location(stmt2, gimple_location(stmt));
				wi->gsi = gsi_for_stmt(stmt);
				gsi_replace(&wi->gsi, stmt2, false);
			}

			break;
		}

		case PARM_DECL:
		case VAR_DECL:
			*walk_subtrees = 0;
			if (op0 != oldvar)
				break;

			*tp = newvar;
			if (is_gimple_assign(stmt))
				gimple_assign_set_rhs_code(stmt, TREE_CODE(oldvar));
			else
				gcc_assert(gimple_code(stmt) == GIMPLE_ASM);
			update_stmt(stmt);
			break;
		}
		break;
	}

	case COMPONENT_REF: {
		tree op0;

		op0 = TREE_OPERAND(op, 0);
		if (!VAR_P(op0))
			break;
		if (op0 != oldvar)
			break;

		switch (gimple_code(stmt)) {
		default:
			debug_gimple_stmt(stmt);
			debug_tree(op);
			debug_tree(oldvar);
			gcc_unreachable();

		case GIMPLE_ASM:
		case GIMPLE_ASSIGN:
		case GIMPLE_CALL:
			break;
		}

		gcc_assert(POINTER_TYPE_P(TREE_TYPE(newvar)));
		TREE_OPERAND(op, 0) = build2_loc(DECL_SOURCE_LOCATION(oldvar), MEM_REF, TREE_TYPE(oldvar), newvar, build_int_cst(TREE_TYPE(newvar), 0));
		*walk_subtrees = 0;
		update_stmt(stmt);
		break;
	}

	case CONSTRUCTOR: {
		tree lhs;

		*walk_subtrees = 0;
		gcc_assert(gimple_assign_single_p(stmt));
		lhs = gimple_assign_lhs(stmt);
		if (!VAR_P(lhs))
			break;
		if (lhs != oldvar)
			break;

		debug_gimple_stmt(stmt);
		debug_tree(oldvar);
		debug_tree(gimple_assign_rhs1(stmt));
		debug_tree(TREE_TYPE(gimple_assign_rhs1(stmt)));
		gcc_assert(wi->is_lhs);
		gcc_unreachable();
	}

	case PARM_DECL:
	case VAR_DECL: {
		tree lhs, rhs;
		gimple_stmt_ptr stmt2;
		gimple_stmt_iterator gsi;

		*walk_subtrees = 0;
		if (op != oldvar)
			break;

		switch (gimple_code(stmt)) {
		default:
			debug_gimple_stmt(stmt);
			gcc_unreachable();

		case GIMPLE_ASSIGN:
			gcc_assert(gimple_assign_single_p(stmt));
			if (wi->is_lhs) {
				gcc_assert(gimple_assign_lhs(stmt) == oldvar);
				lhs = build2_loc(DECL_SOURCE_LOCATION(oldvar), MEM_REF, TREE_TYPE(oldvar), newvar, build_int_cst(TREE_TYPE(newvar), 0));
				rhs = gimple_assign_rhs1(stmt);
				stmt2 = gimple_build_assign_with_ops(gimple_assign_rhs_code(stmt), lhs, rhs, NULL_TREE);
				gimple_set_block(stmt2, gimple_block(stmt));
				gimple_set_location(stmt2, gimple_location(stmt));
				gsi = gsi_for_stmt(stmt);
				gsi_insert_before(&gsi, stmt2, GSI_SAME_STMT);
				gsi_remove(&gsi, true);
			} else {
				if (gimple_assign_rhs_code(stmt) == BIT_FIELD_REF) {
					poly_int64_pod bitsize, bitpos;

					rhs = gimple_assign_rhs1(stmt);
					bitsize = TREE_INT_CST_LOW(TREE_OPERAND(rhs, 1));
					gcc_assert(multiple_p(bitsize, BITS_PER_UNIT));
					bitpos = TREE_INT_CST_LOW(TREE_OPERAND(rhs, 2));
					gcc_assert(multiple_p(bitpos, BITS_PER_UNIT));
					rhs = build2_loc(DECL_SOURCE_LOCATION(oldvar), MEM_REF, TREE_TYPE(rhs), newvar, build_int_cst(TREE_TYPE(newvar), exact_div(bitpos, BITS_PER_UNIT)));
				} else {
					gcc_assert(gimple_assign_rhs1(stmt) == oldvar);
					rhs = build2_loc(DECL_SOURCE_LOCATION(oldvar), MEM_REF, TREE_TYPE(oldvar), newvar, build_int_cst(TREE_TYPE(newvar), 0));
				}
				lhs = gimple_assign_lhs(stmt);
				stmt2 = gimple_build_assign_with_ops(gimple_assign_rhs_code(stmt), lhs, rhs, NULL_TREE);
				gimple_set_block(stmt2, gimple_block(stmt));
				gimple_set_location(stmt2, gimple_location(stmt));
				gsi = gsi_for_stmt(stmt);
				gsi_replace(&gsi, stmt2, false);
			}
			break;

		case GIMPLE_CALL:
			gcc_assert(wi->is_lhs);
			gcc_assert(gimple_call_lhs(stmt) == oldvar);
			lhs = build2_loc(DECL_SOURCE_LOCATION(oldvar), MEM_REF, TREE_TYPE(oldvar), newvar, build_int_cst(TREE_TYPE(newvar), 0));
			gimple_call_set_lhs(stmt, lhs);
			update_stmt(stmt);
			break;
		}

		break;
	}
	}

	return NULL_TREE;
}

static void handle_nolocal_var(int *postorder, unsigned int postorder_num, sbitmap var_use_blocks, sbitmap var_exit_blocks, tree oldvar, bitmap nolocalvars)
{
	basic_block var_scope_entry_bb, var_scope_exit_bb;
	unsigned int i;
	unsigned HOST_WIDE_INT size, align;
	tree newvar;
	sbitmap_iterator sbi;
	bool initialized = false, freed = false;
	gimple_stmt_ptr first_use = NULL;

	if (bitmap_bit_p(nolocalvars, DECL_UID(oldvar)))
		return;
	bitmap_set_bit(nolocalvars, DECL_UID(oldvar));

	sbitmap_clear(var_use_blocks);
	sbitmap_clear(var_exit_blocks);

	// 1. find BBs using oldvar
	var_scope_entry_bb = EXIT_BLOCK_PTR_FOR_FN(cfun);
	var_scope_exit_bb = single_succ(ENTRY_BLOCK_PTR_FOR_FN(cfun));

	for (i = postorder_num; i; i--) {
		basic_block bb;
		gimple_stmt_iterator gsi;
		gphi_iterator gsi2;

		bb = BASIC_BLOCK_FOR_FN(cfun, postorder[i - 1]);

		// 1.1 PHIs should not have any nolocals
		for (gsi2 = gsi_start_phis(bb); !gsi_end_p(gsi2); gsi_next(&gsi2)) {
			struct walk_stmt_info wi = {};
			gphi *phi_stmt = gsi_gphi(gsi2);
			unsigned int i, e;

			wi.info = oldvar;
			for (i = 0, e = gimple_phi_num_args(phi_stmt); i < e; i++) {
				if (walk_tree(gimple_phi_arg_def_ptr(phi_stmt, i), uses_nolocal_var, &wi, NULL) == NULL_TREE)
					continue;

				sbitmap_set_bit(var_use_blocks, bb->index);
				if (TREE_CODE(gimple_phi_arg_def(phi_stmt, i)) == ADDR_EXPR)
					continue;

				debug_gimple_stmt(phi_stmt);
				debug_tree(gimple_phi_arg_def(phi_stmt, i));
				gcc_unreachable();
			}
		}

		// 1.2 remember each BB that uses oldvar to speed later walks
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			struct walk_stmt_info wi = {};
			gimple_stmt_ptr stmt = gsi_stmt(gsi);

			wi.info = oldvar;
			if (walk_gimple_stmt(&gsi, NULL, uses_nolocal_var, &wi) == NULL_TREE)
				continue;

			// ignore the fake assignment from auto-var-init
			if (gimple_deferred_init_p(stmt))
				continue;

			if (wi.is_lhs && is_gimple_assign(stmt)) {
				gcc_assert(gimple_assign_single_p(stmt) || gimple_assign_cast_p(stmt));

				if (gimple_clobber_p(stmt)) {
					if (initialized && !dominated_by_p(CDI_DOMINATORS, bb, var_scope_entry_bb)) {
						if (report_nolocals)
							inform(DECL_SOURCE_LOCATION(oldvar), "ignoring bad end-of-scope indicator for variable %qD in BB %u (%u)", oldvar, bb->index, var_scope_entry_bb->index);
						initialized = false;
					}
#if BUILDING_GCC_VERSION >= 12000
					gcc_assert(gimple_clobber_p(stmt, CLOBBER_EOL));
#endif
					sbitmap_set_bit(var_exit_blocks, bb->index);
					var_scope_exit_bb = nearest_common_dominator(CDI_POST_DOMINATORS, var_scope_entry_bb, bb);
					continue;
				}

				if (var_scope_entry_bb == EXIT_BLOCK_PTR_FOR_FN(cfun) && get_inner_arg(gimple_assign_lhs(stmt)) == oldvar && gimple_assign_rhs_code(stmt) == CONSTRUCTOR) {
					gcc_assert(!initialized);
					sbitmap_set_bit(var_use_blocks, bb->index);
					initialized = true;
					first_use = stmt;
					var_scope_entry_bb = bb;
					continue;
				}
			}

			sbitmap_set_bit(var_use_blocks, bb->index);
			if (initialized || first_use) {
				var_scope_exit_bb = nearest_common_dominator(CDI_POST_DOMINATORS, var_scope_entry_bb, bb);
				continue;
			}

			gcc_assert(!is_gimple_assign(stmt) || gimple_assign_lhs(stmt) != oldvar || gimple_assign_rhs_code(stmt) != CONSTRUCTOR);
			first_use = stmt;
			var_scope_entry_bb = bb;
			var_scope_exit_bb = bb;
		}
	}
	if (first_use && initialized)
		first_use = NULL;

	if (VAR_P(oldvar) && DECL_INITIAL(oldvar) != NULL_TREE) {
debug_tree(oldvar);
debug_tree(DECL_INITIAL(oldvar));
gcc_unreachable();
		gcc_assert(-1 == sbitmap_first_set_bit(var_use_blocks));
		gcc_assert(-1 != sbitmap_first_set_bit(var_exit_blocks));
		gcc_assert(!first_use && !initialized && !freed);
		return;
	}

	sbitmap_ior(var_use_blocks, var_use_blocks, var_exit_blocks);
	if (-1 == sbitmap_first_set_bit(var_use_blocks))
		return;

	if (!first_use && !initialized)
		return;

	gcc_assert(!!first_use != !!initialized);

	freed = -1 != sbitmap_first_set_bit(var_exit_blocks);

	switch (TREE_CODE(oldvar)) {
	default:

	case PARM_DECL:
		if (!initialized && !freed)
			break;
		gcc_unreachable();

	case VAR_DECL: {
		bool changed_entry, changed_exit;

		if (initialized || freed)
			break;

		do {
			changed_entry = false;
			while (var_scope_entry_bb != ENTRY_BLOCK_PTR_FOR_FN(cfun) &&
			       !dominated_by_p(CDI_DOMINATORS, var_scope_exit_bb, var_scope_entry_bb))
			{
				var_scope_entry_bb = get_immediate_dominator(CDI_DOMINATORS, var_scope_entry_bb);
				changed_entry = true;
			}

			changed_exit = false;
			while (var_scope_exit_bb != EXIT_BLOCK_PTR_FOR_FN(cfun) &&
			       !dominated_by_p(CDI_POST_DOMINATORS, var_scope_entry_bb, var_scope_exit_bb))
			{
				var_scope_exit_bb = get_immediate_dominator(CDI_POST_DOMINATORS, var_scope_exit_bb);
				changed_exit = true;
			}
		} while (changed_entry || changed_exit);
	}
	}

	size = TREE_INT_CST_LOW(TYPE_SIZE_UNIT(TREE_TYPE(oldvar)));
	align = TYPE_ALIGN_UNIT(TREE_TYPE(oldvar));
	gcc_assert(NOLOCALS_SIZE != NULL_TREE);
	if (size > TREE_INT_CST_LOW(NOLOCALS_SIZE)) {
		error_at(DECL_SOURCE_LOCATION(oldvar), "the size of nolocal variable %qD is larger than " HOST_WIDE_INT_PRINT_UNSIGNED " bytes", oldvar, TREE_INT_CST_LOW(NOLOCALS_SIZE));
		return;
	}
	if (align > TREE_INT_CST_LOW(NOLOCALS_ALIGN)) {
		error_at(DECL_SOURCE_LOCATION(oldvar), "the alignment of nolocal variable %qD is larger than " HOST_WIDE_INT_PRINT_UNSIGNED " bytes", oldvar, TREE_INT_CST_LOW(NOLOCALS_ALIGN));
		return;
	}

	if (report_nolocals)
		inform(DECL_SOURCE_LOCATION(oldvar),
			"converting nolocal variable %u %u %qE",
			(unsigned int)TREE_INT_CST_LOW(TYPE_SIZE_UNIT(TREE_TYPE(oldvar))),
			(unsigned int)TYPE_ALIGN_UNIT(TREE_TYPE(oldvar)),
			oldvar);

	// 2. prepare the off-stack replacement
	newvar = create_tmp_var(build_pointer_type(TREE_TYPE(oldvar)), "nolocal_var");
	DECL_SOURCE_LOCATION(newvar) = DECL_SOURCE_LOCATION(oldvar);
	mark_sym_for_renaming(newvar);

	// 3. move complex expressions involving the nolocal variable to assignments
	EXECUTE_IF_SET_IN_SBITMAP(var_use_blocks, 0, i, sbi) {
		basic_block bb;
		gimple_stmt_iterator gsi;
		gphi_iterator gsi2;

		bb = BASIC_BLOCK_FOR_FN(cfun, i);

		for (gsi2 = gsi_start_phis(bb); !gsi_end_p(gsi2); gsi_next(&gsi2)) {
			struct walk_stmt_info wi = {};
			struct walk_stmt_info wi2 = {};
			struct nolocal_var vars = { oldvar, newvar };
			gphi *phi_stmt = gsi_gphi(gsi2);
			unsigned int i, e;

			wi.info = oldvar;
			wi2.info = &vars;

			for (i = 0, e = gimple_phi_num_args(phi_stmt); i < e; i++) {
				tree newvar_phi, block;
				gimple_stmt_ptr assign;
				location_t loc;
				edge e;

				if (walk_tree(gimple_phi_arg_def_ptr(phi_stmt, i), uses_nolocal_var, &wi, NULL) == NULL_TREE)
					continue;

				e = gimple_phi_arg_edge(phi_stmt, i);
				gcc_assert(e->src != ENTRY_BLOCK_PTR_FOR_FN(cfun));
				sbitmap_set_bit(var_use_blocks, e->src->index);

				gsi = gsi_last_nondebug_bb(e->src);
				if (gsi_end_p(gsi)) {
					loc = UNKNOWN_LOCATION;
					block = NULL_TREE;
				} else {
					loc = gimple_location(gsi_stmt(gsi));
					block = gimple_block(gsi_stmt(gsi));
				}
				if (loc == UNKNOWN_LOCATION)
					loc = DECL_SOURCE_LOCATION(newvar);

				newvar_phi = create_tmp_var(TREE_TYPE(newvar), "nolocal_temp_phi");
				newvar_phi = make_ssa_name(newvar_phi, NULL);
				assign = gimple_build_assign(newvar_phi, unshare_expr(gimple_phi_arg_def(phi_stmt, i)));
				SSA_NAME_DEF_STMT(newvar_phi) = assign;
				gimple_set_block(assign, block);
				gimple_set_location(assign, loc);
				gsi_insert_before(&gsi, assign, GSI_NEW_STMT);

				wi2.gsi = gsi;
				walk_tree(gimple_op_ptr(assign, 1), convert_nolocal_var, &wi2, NULL);

				add_phi_arg(phi_stmt, gimple_assign_lhs(assign), e, loc);
				update_stmt(phi_stmt);
			}
		}

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi);) {
			tree rhs;
			gimple_stmt_ptr stmt, assign;
			struct walk_stmt_info wi = {};

			wi.info = oldvar;
			stmt = gsi_stmt(gsi);
			switch (gimple_code(stmt)) {
			default:
				if (walk_gimple_stmt(&gsi, NULL, uses_nolocal_var, &wi) == NULL_TREE)
					break;
				debug_gimple_stmt(stmt);
				debug_tree(oldvar);
				gcc_unreachable();

			case GIMPLE_ASM:
				break;

			case GIMPLE_ASSIGN:
				if (walk_tree(gimple_op_ptr(stmt, 0), uses_nolocal_var, &wi, NULL) != NULL_TREE) {
					sbitmap_set_bit(var_use_blocks, bb->index);

					gcc_assert(gimple_assign_single_p(stmt) || gimple_assign_cast_p(stmt));

					// we don't need the fake assignment from auto-var-init
					if (gimple_deferred_init_p(stmt)) {
						gsi_remove(&gsi, true);
						continue;
					}

					// free 'allocated' nolocal replacement
					if (gimple_clobber_p(stmt)) {
						if (initialized) {
							gimple_stmt_ptr dtor;

							gcc_assert(free_nolocal_decl);
							dtor = gimple_build_call(free_nolocal_decl, 2, newvar, TYPE_SIZE_UNIT(TREE_TYPE(oldvar)));
							gimple_set_block(dtor, gimple_block(stmt));
							gimple_set_location(dtor, gimple_location(stmt));
							gsi_insert_before(&gsi, dtor, GSI_SAME_STMT);
						}

						gsi_remove(&gsi, true);
						continue;
					}

					// allocate or initialize nolocal replacement
					// allocate_nolocal will always zero-init but we could also pass a 'need zero init' flag
					if (get_inner_arg(gimple_assign_lhs(stmt)) == oldvar && gimple_assign_rhs_code(stmt) == CONSTRUCTOR) {
						gimple_stmt_ptr ctor;

						gcc_assert(!CONSTRUCTOR_NELTS(gimple_assign_rhs1(stmt)));
						gcc_assert(alloc_nolocal_decl);

						if (!initialized) {
							gsi_remove(&gsi, true);
							continue;
						}

						if (freed)
							ctor = gimple_build_call(alloc_nolocal_decl, 2, TYPE_SIZE_UNIT(TREE_TYPE(oldvar)), size_int(TYPE_ALIGN_UNIT(TREE_TYPE(oldvar))));
						else
							ctor = gimple_build_call(builtin_decl_implicit(BUILT_IN_MEMSET), 3, newvar, integer_zero_node, TYPE_SIZE_UNIT(TREE_TYPE(oldvar)));
						gimple_call_set_lhs(ctor, newvar);
						gimple_set_block(ctor, gimple_block(stmt));
						gimple_set_location(ctor, gimple_location(stmt));
						gsi_insert_before(&gsi, ctor, GSI_SAME_STMT);

						gsi_remove(&gsi, true);
						continue;
					}
				}
				// FALLTHROUGH

			case GIMPLE_CALL:
				if (is_gimple_call(stmt) && walk_tree(gimple_op_ptr(stmt, 0), uses_nolocal_var, &wi, NULL) != NULL_TREE) {
					// we don't need the fake assignment from auto-var-init
					if (gimple_deferred_init_p(stmt)) {
						gsi_remove(&gsi, true);
						continue;
					}
				}
				// FALLTHROUGH

			case GIMPLE_COND: {
				unsigned int i, e;

				i = gimple_code(stmt) == GIMPLE_COND || gimple_code(stmt) == GIMPLE_ASM ? 0 : 1;
				e = gimple_code(stmt) == GIMPLE_COND ? 2 : gimple_num_ops(stmt);
				for (; i < e; i++) {
					tree lhs;

					if (walk_tree(gimple_op_ptr(stmt, i), uses_nolocal_var, &wi, NULL) == NULL_TREE)
						continue;

					gcc_assert(sbitmap_bit_p(var_use_blocks, bb->index));

					rhs = unshare_expr(gimple_op(stmt, i));
					if (gimple_code(stmt) == GIMPLE_ASM) {
						debug_gimple_stmt(stmt);
						debug_tree(rhs);
						rhs = TREE_VALUE(rhs);
						debug_tree(rhs);
						gcc_unreachable();
					}

					lhs = create_tmp_var(TREE_TYPE(rhs), "nolocal_temp");
					add_referenced_var(lhs);
					mark_sym_for_renaming(lhs);

					assign = gimple_build_assign(lhs, rhs);
					gimple_set_block(assign, gimple_block(stmt));
					gimple_set_location(assign, gimple_location(stmt));
					gsi_insert_before(&gsi, assign, GSI_SAME_STMT);
					gcc_assert(TREE_CODE(lhs) != SSA_NAME);

					if (gimple_code(stmt) == GIMPLE_ASM)
						TREE_VALUE(gimple_op(stmt, i)) = lhs;
					else
						*gimple_op_ptr(stmt, i) = lhs;
					update_stmt(stmt);
					if (first_use == stmt)
						first_use = assign;
				}

				break;
			}

			case GIMPLE_DEBUG:
				gsi_remove(&gsi, true);
				continue;
			}

			gsi_next(&gsi);
		}
	}

	// 4. convert the nolocals
	EXECUTE_IF_SET_IN_SBITMAP(var_use_blocks, 0, i, sbi) {
		basic_block bb;
		gimple_stmt_iterator gsi;

		bb = BASIC_BLOCK_FOR_FN(cfun, i);

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			struct walk_stmt_info wi = {};
			struct nolocal_var vars = { oldvar, newvar };
			gimple_stmt_ptr stmt = gsi_stmt(gsi);

			wi.info = &vars;
			wi.gsi = gsi;

			switch (gimple_code(stmt)) {
			default:
				continue;

			case GIMPLE_ASM: {
				unsigned int e;

				wi.is_lhs = 0;
				for (i = 0, e = gimple_num_ops(stmt); i < e; i++)
					walk_tree(&TREE_VALUE(gimple_op(stmt, i)), convert_nolocal_var, &wi, NULL);

				continue;
			}

			case GIMPLE_ASSIGN:
			case GIMPLE_CALL:
				wi.is_lhs = 1;
				walk_tree(gimple_op_ptr(stmt, 0), convert_nolocal_var, &wi, NULL);
				// FALLTHROUGH

			case GIMPLE_COND: {
				unsigned int i, e;

				i = gimple_code(stmt) == GIMPLE_COND ? 0 : 1;
				e = gimple_code(stmt) == GIMPLE_COND ? 2 : gimple_num_ops(stmt);
				wi.is_lhs = 0;
				for (; i < e; i++)
					walk_tree(gimple_op_ptr(stmt, i), convert_nolocal_var, &wi, NULL);
			}
			}
		}
	}

	// 5. handle missing ctor/dtor cases
	if (!initialized || !freed) {
		gimple_stmt_iterator gsi;
		gimple_stmt_ptr ctor, stmt = NULL;
		location_t loc;

		loc = DECL_SOURCE_LOCATION(oldvar);

		gsi = gsi_start_nondebug_after_labels_bb(single_succ(ENTRY_BLOCK_PTR_FOR_FN(cfun)));
		gcc_assert(alloc_nolocal_decl);
		ctor = gimple_build_call(alloc_nolocal_decl, 2, TYPE_SIZE_UNIT(TREE_TYPE(oldvar)), size_int(TYPE_ALIGN_UNIT(TREE_TYPE(oldvar))));
		gimple_set_location(ctor, loc);
		if (!gsi_end_p(gsi)) {
			stmt = gsi_stmt(gsi);
			if (stmt)
				gimple_set_block(ctor, gimple_block(stmt));
		}
		gimple_call_set_lhs(ctor, newvar);
		gsi_insert_before(&gsi, ctor, GSI_SAME_STMT);

		if (TREE_CODE(oldvar) == PARM_DECL) {
			tree lhs;
			gimple_stmt_ptr stmt2;

			lhs = build2_loc(loc, MEM_REF, TREE_TYPE(oldvar), newvar, build_int_cst(TREE_TYPE(newvar), 0));
			stmt2 = gimple_build_assign(lhs, oldvar);
			gimple_set_location(stmt2, loc);
			if (stmt)
				gimple_set_block(stmt2, gimple_block(stmt));
			gsi_insert_before(&gsi, stmt2, GSI_SAME_STMT);
		}
	}

	if ((!initialized || !freed) && EDGE_COUNT(EXIT_BLOCK_PTR_FOR_FN(cfun)->preds)) {
		gimple_stmt_iterator gsi;
		gimple_stmt_ptr dtor, stmt;

		gcc_assert(single_pred_p(EXIT_BLOCK_PTR_FOR_FN(cfun)));
		gsi = gsi_last_nondebug_bb(single_pred(EXIT_BLOCK_PTR_FOR_FN(cfun)));

		gcc_assert(free_nolocal_decl);

		dtor = gimple_build_call(free_nolocal_decl, 2, newvar, TYPE_SIZE_UNIT(TREE_TYPE(oldvar)));
		stmt = gsi_stmt(gsi);
		if (stmt) {
			gimple_set_block(dtor, gimple_block(stmt));
			gimple_set_location(dtor, gimple_location(stmt));
		}
		gsi_insert_before(&gsi, dtor, GSI_SAME_STMT);
	}

	// 6. verify that no nolocal variable was left unconverted
	EXECUTE_IF_SET_IN_SBITMAP(var_use_blocks, 0, i, sbi) {
		basic_block bb;
		gimple_stmt_iterator gsi;

		bb = BASIC_BLOCK_FOR_FN(cfun, i);
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;
			struct walk_stmt_info wi = {};

			wi.info = oldvar;
			stmt = gsi_stmt(gsi);
			if (walk_gimple_stmt(&gsi, NULL, uses_nolocal_var, &wi) == NULL_TREE)
				continue;

			if (TREE_CODE(oldvar) == PARM_DECL && !wi.is_lhs) {
				gimple_stmt_iterator gsi2;
				gimple_stmt_ptr stmt2;

				gsi2 = gsi_for_stmt(stmt);
				gsi_prev(&gsi2);
				gcc_assert(!gsi_end_p(gsi2));
				stmt2 = gsi_stmt(gsi2);
				if (gimple_code(stmt2) == GIMPLE_CALL) {
					tree fndecl;

					fndecl = gimple_call_fndecl(stmt2);
					if (fndecl == alloc_nolocal_decl)
						continue;
				}
			}

			debug_gimple_stmt(stmt);
			debug_tree(oldvar);
			gcc_unreachable();
		}
	}
}

static void convert_nolocal_vars(int *postorder, unsigned int postorder_num, sbitmap var_use_blocks, sbitmap var_exit_blocks, bitmap nolocalvars)
{
	unsigned int i;
	tree oldvar;

	FOR_EACH_LOCAL_DECL(cfun, i, oldvar) {
		location_t loc;
		unsigned HOST_WIDE_INT size, align;

		gcc_assert(VAR_P(oldvar));
		if (!is_nolocal_var(oldvar))
			continue;

		loc = DECL_SOURCE_LOCATION(oldvar);
		size = TREE_INT_CST_LOW(TYPE_SIZE_UNIT(TREE_TYPE(oldvar)));
		align = TYPE_ALIGN_UNIT(TREE_TYPE(oldvar));
		gcc_assert(NOLOCALS_SIZE != NULL_TREE);
		if (size > TREE_INT_CST_LOW(NOLOCALS_SIZE)) {
			error_at(loc, "the size of nolocal variable %qD is larger than " HOST_WIDE_INT_PRINT_UNSIGNED " bytes", oldvar, TREE_INT_CST_LOW(NOLOCALS_SIZE));
			continue;
		}
		if (align > TREE_INT_CST_LOW(NOLOCALS_ALIGN)) {
			error_at(loc, "the alignment of nolocal variable %qD is larger than " HOST_WIDE_INT_PRINT_UNSIGNED " bytes", oldvar, TREE_INT_CST_LOW(NOLOCALS_ALIGN));
			continue;
		}

		handle_nolocal_var(postorder, postorder_num, var_use_blocks, var_exit_blocks, oldvar, nolocalvars);
	}
}

static tree translate_argnum(gimple_stmt_ptr stmt, unsigned int argnum)
{
	cgraph_node_ptr node;
	tree fndecl, orig_decl, decl_list, orig_decl_list;
	tree orig_arg;

	fndecl = gimple_call_fndecl(stmt);
	if (fndecl == NULL_TREE)
		return gimple_call_arg(stmt, argnum);

	node = cgraph_get_node(fndecl);
	gcc_assert(node);

	if (node->clone_of && NODE_CLONE_INFO(node) && NODE_CLONE_INFO(node)->tree_map)
#if BUILDING_GCC_VERSION >= 10000
		gcc_assert(!NODE_CLONE_INFO(node)->param_adjustments);
#else
		gcc_assert(!node->clone.args_to_skip);
#endif

	if (!DECL_ARTIFICIAL(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) == NULL_TREE)
		return gimple_call_arg(stmt, argnum);

	orig_decl = DECL_ABSTRACT_ORIGIN(fndecl);
	gcc_assert(orig_decl != NULL_TREE);

	decl_list = DECL_ARGUMENTS(fndecl);
	orig_decl_list = DECL_ARGUMENTS(orig_decl);
	gcc_assert(decl_list != NULL_TREE && orig_decl_list != NULL_TREE);

	if (list_length(decl_list) == list_length(orig_decl_list))
		return gimple_call_arg(stmt, argnum);

	orig_arg = chain_index(argnum, orig_decl_list);
	gcc_assert(orig_arg != NULL_TREE);

	for (argnum = 0; decl_list != NULL_TREE; argnum++, decl_list = TREE_CHAIN(decl_list)) {
		if (strcmp(DECL_NAME_POINTER(decl_list), DECL_NAME_POINTER(orig_arg)))
			continue;

		return gimple_call_arg(stmt, argnum);
	}

	return NULL_TREE;
}

static void convert_nolocal_args(int *postorder, unsigned int postorder_num, sbitmap var_use_blocks, sbitmap var_exit_blocks, bitmap nolocalvars)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;
			tree attr;

			stmt = gsi_stmt(gsi);
			if (gimple_code(stmt) != GIMPLE_CALL)
				continue;

			attr = get_nolocal_attr(stmt);
			if (attr == NULL_TREE)
				continue;

			for (attr = TREE_VALUE(attr); attr; attr = TREE_CHAIN(attr)) {
				tree id, oldvar, basevar;
				unsigned int argnum;

				id = TREE_VALUE(attr);
				gcc_assert(TREE_CODE(id) == INTEGER_CST);
				argnum = TREE_INT_CST_LOW(id) - 1;
				oldvar = translate_argnum(stmt, argnum);
				if (oldvar == NULL_TREE)
					continue;

				switch (TREE_CODE(oldvar)) {
				default:
					debug_gimple_stmt(stmt);
					debug_tree(oldvar);
					gcc_unreachable();

				case INTEGER_CST:
				case PARM_DECL:
				case VAR_DECL:
					continue;

				case SSA_NAME: {
					gimple_stmt_ptr def_stmt;
					def_stmt = SSA_NAME_DEF_STMT(oldvar);
					if (!is_gimple_assign(def_stmt))
						continue;
					if (gimple_assign_rhs_code(def_stmt) != ADDR_EXPR)
						continue;

					oldvar = gimple_assign_rhs1(def_stmt);
				}
				// FALLTHROUGH

				case ADDR_EXPR:
					basevar = oldvar;
					oldvar = get_inner_arg(TREE_OPERAND(oldvar, 0));
					if (TREE_STATIC(oldvar))
						continue;

					switch (TREE_CODE(oldvar)) {
					default:
						debug_gimple_stmt(stmt);
						debug_tree(basevar);
						debug_tree(oldvar);
						gcc_unreachable();

#if BUILDING_GCC_VERSION >= 4006
					case MEM_REF:
						continue;
#endif

					case PARM_DECL:
					case VAR_DECL:
						if (DECL_CONTEXT(oldvar) != current_function_decl)
							continue;
						break;
					}
					break;
				}

				handle_nolocal_var(postorder, postorder_num, var_use_blocks, var_exit_blocks, oldvar, nolocalvars);
			}
		}
	}
}

static bool propagate_nolocal_types(tree type, bitmap nolocaltypes)
{
	tree field;
	bool nolocal = false;

	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		tree fieldtype;

		fieldtype = strip_array_types(TREE_TYPE(field));

		if (RECORD_OR_UNION_TYPE_P(fieldtype)) {
			if (bitmap_bit_p(nolocaltypes, TYPE_UID(TYPE_CANONICAL(fieldtype))))
				continue;

			bitmap_set_bit(nolocaltypes, TYPE_UID(TYPE_CANONICAL(fieldtype)));
			nolocal |= propagate_nolocal_types(fieldtype, nolocaltypes);
			continue;
		}

		if (POINTER_TYPE_P(fieldtype)) {
			fieldtype = strip_array_types(TREE_TYPE(fieldtype));
			if (!RECORD_OR_UNION_TYPE_P(fieldtype))
				continue;
			if (lookup_attribute("nolocal", TYPE_ATTRIBUTES(fieldtype)))
				continue;

			gcc_assert(TYPE_CANONICAL(fieldtype));
			bitmap_set_bit(nolocaltypes, TYPE_UID(TYPE_CANONICAL(fieldtype)));
			nolocal = true;
		}
	}

	return nolocal;
}

static void propagate_nolocal_vars(int *postorder, unsigned int postorder_num, sbitmap var_use_blocks, sbitmap var_exit_blocks, bitmap nolocalvars)
{
	unsigned int i;
	tree oldvar;
	bitmap nolocaltypes;
	basic_block bb;

	nolocaltypes = BITMAP_ALLOC(NULL);
	bitmap_clear(nolocaltypes);

	// mark pointed to types in nolocal vars as nolocal
	FOR_EACH_LOCAL_DECL(cfun, i, oldvar) {
		tree type;

		type = strip_array_types(TREE_TYPE(oldvar));
		if (!RECORD_OR_UNION_TYPE_P(type))
			continue;
		if (!bitmap_bit_p(nolocalvars, DECL_UID(oldvar)))
			continue;

		propagate_nolocal_types(type, nolocaltypes);
	}

	// handle newly exposed nolocal vars
	FOR_EACH_LOCAL_DECL(cfun, i, oldvar) {
		tree type;

		if (TREE_STATIC(oldvar))
			continue;
		if (!TREE_ADDRESSABLE(oldvar))
			continue;
		if (bitmap_bit_p(nolocalvars, DECL_UID(oldvar)))
			continue;

		type = strip_array_types(TREE_TYPE(oldvar));
		if (!RECORD_OR_UNION_TYPE_P(type))
			continue;
		gcc_assert(TYPE_CANONICAL(type));
		if (!bitmap_bit_p(nolocaltypes, TYPE_UID(TYPE_CANONICAL(type))))
			continue;

		handle_nolocal_var(postorder, postorder_num, var_use_blocks, var_exit_blocks, oldvar, nolocalvars);
	}

	// handle nolocal vars whose address escapes into an existing one
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;
			tree lhs;

			stmt = gsi_stmt(gsi);
			if (!is_gimple_assign(stmt))
				continue;
			if (gimple_assign_rhs_code(stmt) != ADDR_EXPR)
				continue;

			lhs = gimple_assign_lhs(stmt);
			lhs = get_inner_arg(lhs);
			if (!TYPE_CANONICAL(TREE_TYPE(lhs)))
				continue;
			if (!bitmap_bit_p(nolocaltypes, TYPE_UID(TYPE_CANONICAL(TREE_TYPE(lhs)))))
				continue;

			oldvar = gimple_assign_rhs1(stmt);
			oldvar = get_inner_arg(TREE_OPERAND(oldvar, 0));
			if (DECL_EXTERNAL(oldvar))
				continue;
			if (TREE_STATIC(oldvar))
				continue;
			gcc_assert(TREE_ADDRESSABLE(oldvar));

			handle_nolocal_var(postorder, postorder_num, var_use_blocks, var_exit_blocks, oldvar, nolocalvars);
		}
	}

	BITMAP_FREE(nolocaltypes);
}

unsigned int convert_nolocal_variables_execute(void)
{
	int *postorder;
	unsigned int postorder_num;
	sbitmap var_use_blocks, var_exit_blocks;
	bitmap nolocalvars;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	get_single_pred_entry_block();

	var_use_blocks = sbitmap_alloc(last_basic_block_for_fn(cfun));
	var_exit_blocks = sbitmap_alloc(last_basic_block_for_fn(cfun));

	nolocalvars = BITMAP_ALLOC(NULL);
	bitmap_clear(nolocalvars);

	postorder = XNEWVEC(int, n_basic_blocks_for_fn(cfun));
	postorder_num = post_order_compute(postorder, true, false);
	gcc_assert(ENTRY_BLOCK == postorder[postorder_num - 1]);

	convert_nolocal_vars(postorder, postorder_num, var_use_blocks, var_exit_blocks, nolocalvars);
	convert_nolocal_args(postorder, postorder_num, var_use_blocks, var_exit_blocks, nolocalvars);
	propagate_nolocal_vars(postorder, postorder_num, var_use_blocks, var_exit_blocks, nolocalvars);

	XDELETEVEC(postorder);
	BITMAP_FREE(nolocalvars);
	sbitmap_free(var_exit_blocks);
	sbitmap_free(var_use_blocks);
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

#define PASS_NAME convert_nolocal_variables
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_START TODO_remove_unused_locals
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_verify_flow | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"
