/*
 * Copyright 2022 by Open Source Security, Inc., PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * replaces the i386 regparm(3) calling convention with regparm(2) using edx/ecx instead of eax/edx
 * this is necessary to free up eax for emitting the SLS protection for calls
 */

#include "utilities.h"

#if defined(TARGET_386) && BUILDING_GCC_VERSION >= 4006

static __typeof__(&default_function_arg) old_function_arg;

static bool should_override_reg(enum machine_mode mode, const_tree type, int orig_nregs)
{
	switch (mode) {
	default:
		return false;
	case E_BLKmode:
	{
		HOST_WIDE_INT bytes;
		bytes = int_size_in_bytes(type);
		if (bytes < 0)
			return false;
		if (CEIL(bytes, UNITS_PER_WORD) <= orig_nregs)
			return true;
		return false;
	}
	case E_DImode:
	case E_SImode:
	case E_HImode:
	case E_QImode:
		if (CEIL(GET_MODE_SIZE(mode), UNITS_PER_WORD) <= orig_nregs)
			return true;

		return false;
	}

	return false;
}

static rtx utilities_function_arg_body(CUMULATIVE_ARGS *args, enum machine_mode mode, const_tree type, rtx reg)
{
	int regno = args->regno;
	int nregs = args->nregs;

	switch (nregs) {
	default:
		gcc_unreachable();
		break;
	case 0:
		gcc_assert(regno == 0);
		break;
	case 1:
		gcc_assert(regno == CX_REG);
		if (should_override_reg(mode, type, nregs))
			reg = NULL_RTX;
		break;
	case 2:
		gcc_assert(regno == DX_REG);
		if (should_override_reg(mode, type, nregs))
			reg = gen_rtx_REG(mode, CX_REG);
		break;
	case 3:
		gcc_assert(regno == AX_REG);
		if (should_override_reg(mode, type, nregs))
			reg = gen_rtx_REG(mode, DX_REG);
		break;
	}

	return reg;
}

#if BUILDING_GCC_VERSION >= 10000
static rtx utilities_function_arg(cumulative_args_t ca, const function_arg_info &fi)
{
	CUMULATIVE_ARGS *args = get_cumulative_args(ca);
	rtx reg = old_function_arg(ca, fi);
	return utilities_function_arg_body(args, fi.mode, fi.type, reg);
}
#elif BUILDING_GCC_VERSION >= 4007
static rtx utilities_function_arg(cumulative_args_t ca, enum machine_mode mode, const_tree type, bool named)
{
	CUMULATIVE_ARGS *args = get_cumulative_args(ca);
	rtx reg = old_function_arg(ca, mode, type, named);
	return utilities_function_arg_body(args, mode, type, reg);
}
#else
static rtx utilities_function_arg(CUMULATIVE_ARGS *args, enum machine_mode mode, const_tree type, bool named)
{
	rtx reg = old_function_arg(args, mode, type, named);
	return utilities_function_arg_body(args, mode, type, reg);
}
#endif

void regparm_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	old_function_arg = targetm.calls.function_arg;

	/*
	 * On i386, function_incoming_arg is the same function as function_arg,
	 * so we don't need to save off/duplicate functions
	 */
	gcc_assert(targetm.calls.function_arg == targetm.calls.function_incoming_arg);

	targetm.calls.function_arg          = utilities_function_arg;
	targetm.calls.function_incoming_arg = utilities_function_arg;
}

#endif
